/************************************************************************
 * E.S.O. - VLT project
 *
 * who       when      what
 * --------  --------  ----------------------------------------------
 * gduvert   2007-05-30 Changed sign of Piston with PHASOR method to have 
 *                      all signs consistent with the _wrong_ sign of the
 *                      PHASE method...
 * gduvert   2006-10-10 Removed selection of two 'best' baseline when enforcing
 *                      piston closure. Do not work with low SNR. 
 * gluck     2006-07-04 Added protected functions for display (tat test):
 *                      amdlibDisplayPiston
 * gduvert   2006-07-04 added amdlibComputePistonIterative2T function and used it
 *                     instead of previous method for amdlib_ITERATIVE_PHASOR.
 * ealtariba 2006-05-24 added treatment for frame selection.
 *                      Corrected errors in piston formulae.
 * ealtariba 2006-04-05 updated according to the new amdlibPISTON and
 *                      amdlibVIS_TABLE_ENTRY structures
 * ealtariba 2006-02-24 differentiated piston per band, implemented function to
 *                      merge and append structures.
 * gzins     26/01/06   Removed debug printf
 * altariba  21/12/05   Created
 */

/**
 * @file
 * Useful functions related to piston computation.
 */

#define _POSIX_SOURCE 1

static char *rcsId="@(#) $Id: amdlibPiston.c,v 1.5 2011/01/24 12:43:53 agabasch Exp $";
static void *use_rcsId = (0 ? &use_rcsId : (void*)&rcsId);

/*
 * System Headers
 */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include "fitsio.h"

/*
 * Local Headers
 */
#include "amdlib.h"
#include "amdlibProtected.h"


#define amdlibMAX_PISTON2_ITERATIONS 50
#define amdlibPRECISION 1E-4

/*
 * Local function declaration
 */
static void amdlibFreePiston(amdlibPISTON *piston);
void amdlibComputePiston2T(int                    nbWlen,
                           float                  *Wave,
                           amdlibCOMPLEX          *x,
                           amdlibCOMPLEX          *sx,
                           float                  *pistonOPD,
                           float                  *sigma,
                           amdlibPISTON_ALGORITHM algoType);
void amdlibComputePistonIterative2T(int                    nbLVis,
                                    float                  *wlen,
                                    amdlibCOMPLEX          *cpxVisTable,
                                    amdlibCOMPLEX          *sigma2CpxVisTable,
                                    float                  *pistonOPD,
                                    float                  *sigma);
void amdlibComputePiston3T(int                    nbWlen,
                           float                  *Wave,
                           amdlibCOMPLEX          *x12,
                           amdlibCOMPLEX          *x23, 
                           amdlibCOMPLEX          *x31,
                           amdlibCOMPLEX          *sx12,
                           amdlibCOMPLEX          *sx23, 
                           amdlibCOMPLEX          *sx31,
                           float                  *pistonOPD,
                           float                  *sigma,
                           amdlibPISTON_ALGORITHM algoType);
/*
 * Public functions
 */
/**
 * Release memory allocated to store pistons, and reset the structure members
 * to zero.
 *
 * @param opd pointer to amdlibPISTON structure.
 */
void amdlibReleasePiston(amdlibPISTON *opd)
{
    amdlibLogTrace("amdlibReleasePiston()");
    
    amdlibFreePiston(opd);
    memset (opd, '\0', sizeof(amdlibPISTON));
}

/*
 * Protected functions
 */
/**
 * Allocate memory for storing pistons. 
 *
 * @param piston pointer to amdlibPISTON structure.
 * @param nbFrames number of frames 
 * @param nbBases number of baselines 
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibAllocatePiston(amdlibPISTON *piston,
                                      const int    nbFrames,
                                      const int    nbBases)
{
    int band;
    
    amdlibLogTrace("amdlibAllocatePiston()");
    
    /* First free previous allocated memory */
    if (piston->thisPtr == piston)
    {
        amdlibFreePiston(piston);
    }

    /* Init data structure */
    piston->thisPtr = memset(piston, '\0', sizeof(*piston));
    /* Set array size */
    piston->nbFrames = nbFrames;
    piston->nbBases  = nbBases;
    for (band = amdlibJ_BAND; band <= amdlibK_BAND; band++)
    {
        piston->bandFlag[band] = amdlibFALSE;
        piston->pistonOPDArray[band] = calloc(nbFrames*nbBases, sizeof(float));
        piston->sigmaPistonArray[band] = 
            calloc(nbFrames*nbBases, sizeof(float));
        if ((piston->pistonOPDArray[band] == NULL) ||
            (piston->sigmaPistonArray[band] == NULL))
        {
            amdlibFreePiston(piston);
            return amdlibFAILURE;
        }
    }
    piston->pistonOPD = calloc(nbFrames*nbBases, sizeof(float));
    piston->sigmaPiston = calloc(nbFrames*nbBases, sizeof(float));
    if ((piston->pistonOPD == NULL) || (piston->sigmaPiston == NULL))
    {
        amdlibFreePiston(piston);
        return amdlibFAILURE;
    }
 
    /* Return */
    return amdlibSUCCESS;
}

/**
 * Append amdlibPISTON data structure. 
 *
 * @param srcOpd pointer to source data structure 
 * @param dstOpd pointer to destination data structure  (append to) 
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibAppendPiston(amdlibPISTON    *dstOpd, 
                                    amdlibPISTON    *srcOpd, 
                                    amdlibERROR_MSG errMsg)
{
    int dentry;
    int sentry;
    int band;

    /* Perform simple check */
    if (dstOpd->nbBases != srcOpd->nbBases)
    {
        amdlibSetErrMsg("Different number of bases");
        return amdlibFAILURE;
    }

    /* Reallocate memory to store additional data */
    for (band = amdlibJ_BAND; band <= amdlibK_BAND; band++)
    {
        dstOpd->pistonOPDArray[band] = realloc(dstOpd->pistonOPDArray[band], 
            (dstOpd->nbFrames + srcOpd->nbFrames) * 
            srcOpd->nbBases * sizeof(float));
        dstOpd->sigmaPistonArray[band] = realloc(dstOpd->sigmaPistonArray[band],            
            (dstOpd->nbFrames + srcOpd->nbFrames) * 
            srcOpd->nbBases * sizeof(float));
        if ((dstOpd->pistonOPDArray[band] == NULL) ||
            (dstOpd->sigmaPistonArray[band] == NULL))
        {
            amdlibSetErrMsg("Could not reallocate memory for piston structure");
            return amdlibFAILURE;
        }
    }
    dstOpd->pistonOPD = realloc(dstOpd->pistonOPD, 
                    (dstOpd->nbFrames + srcOpd->nbFrames) * 
                    srcOpd->nbBases * sizeof(float));
    dstOpd->sigmaPiston = realloc(dstOpd->sigmaPiston,
                    (dstOpd->nbFrames + srcOpd->nbFrames) * 
                    srcOpd->nbBases * sizeof(float));
    if ((dstOpd->pistonOPD == NULL) || (dstOpd->sigmaPiston == NULL))
    {
        amdlibSetErrMsg("Could not reallocate memory for piston structure");
        return amdlibFAILURE;
    }
    
    /* Append fields allocated dynamically */
    dentry = dstOpd->nbFrames * dstOpd->nbBases;
    for (sentry = 0; sentry < srcOpd->nbFrames * srcOpd->nbBases; sentry++)
    {
        for (band=amdlibJ_BAND; band <= amdlibK_BAND; band++)
        {
            dstOpd->pistonOPDArray[band][dentry] = 
                srcOpd->pistonOPDArray[band][sentry];
            dstOpd->sigmaPistonArray[band][dentry] = 
                srcOpd->sigmaPistonArray[band][sentry];
        }
        dstOpd->pistonOPD[dentry] = srcOpd->pistonOPD[sentry];
        dstOpd->sigmaPiston[dentry] = srcOpd->sigmaPiston[sentry];
        dentry++;
    }
    
    /* Append fixed size fields */
    dstOpd->nbFrames += srcOpd->nbFrames;
    return amdlibSUCCESS;
}

/**); 
 * Merge amdlibPISTON data structure. 
 *
 * @param opd1 pointer on merged piston structure.
 * @param opd2 pointer on piston structure to merge.
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibMergePiston(amdlibPISTON    *opd1, 
                                   amdlibPISTON    *opd2, 
                                   amdlibERROR_MSG errMsg)
{
    int band;
    
    memcpy(opd1->pistonOPD, opd2->pistonOPD, 
           opd2->nbFrames * opd2->nbBases * sizeof(float));
    memcpy(opd1->sigmaPiston, opd2->sigmaPiston, 
           opd2->nbFrames * opd2->nbBases * sizeof(float));
    for (band=amdlibJ_BAND; band <= amdlibK_BAND; band++)
    {
        if (opd2->bandFlag[band] == amdlibTRUE)
        {
            if (opd1->bandFlag[band] == amdlibTRUE)
            {
                amdlibSetErrMsg("Same information in pistons to merge");
                return amdlibFAILURE;
            }
            opd1->bandFlag[band] = amdlibTRUE;
            memcpy(opd1->pistonOPDArray[band], opd2->pistonOPDArray[band],
                   opd1->nbFrames * opd1->nbBases * sizeof(float));
            memcpy(opd1->sigmaPistonArray[band], opd2->sigmaPistonArray[band],
                   opd1->nbFrames * opd1->nbBases * sizeof(float));
            return amdlibSUCCESS;
        }
    }
    
    amdlibSetErrMsg("Problem whith opd to merge: doesn't contain value");
    return amdlibFAILURE;
}

/**
 * Split amdlibPISTON input data structure into 3 structures, 1 per spectral
 * band. Each new structure only contains information relative to its associated
 * spectral band .
 *
 * @param srcOpd pointer on piston structure to split.
 * @param dstOpd array of splitted piston structures.
 * @param nbWlen array containing the number of wavelengths per spectral band.
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibSplitPiston(amdlibPISTON    *srcOpd,
                                   amdlibPISTON    *dstOpd,
                                   int             *nbWlen,
                                   amdlibERROR_MSG errMsg)
{
    int band, i;

    for (band=amdlibJ_BAND; band <= amdlibK_BAND; band++)
    {
        if (nbWlen[band] == 0)
        {
            dstOpd[band].thisPtr = NULL;
        }
        else
        {
            /* Allocate memory for dstOpd */
            if (amdlibAllocatePiston(&dstOpd[band], srcOpd->nbFrames, 
                                     srcOpd->nbBases) != amdlibSUCCESS)
            {
                amdlibSetErrMsg("Could not allocate memory for new piston "
                                "structure");
                return amdlibFAILURE;
            }
            
            for (i=amdlibJ_BAND; i <= amdlibK_BAND; i++)
            {
                if (i == band)
                {
                    dstOpd[band].bandFlag[band] = amdlibTRUE;
                }
                else
                {
                    dstOpd[band].bandFlag[i] = amdlibFALSE;
                }
            }
            
            /* Copy fields allocated dynamically */
            memcpy (dstOpd[band].pistonOPDArray[band], 
                    srcOpd->pistonOPDArray[band], 
                    srcOpd->nbFrames * srcOpd->nbBases * sizeof(float));
            memcpy (dstOpd[band].sigmaPistonArray[band], 
                    srcOpd->sigmaPistonArray[band], 
                    srcOpd->nbFrames * srcOpd->nbBases * sizeof(float));

            memcpy (dstOpd[band].pistonOPD, srcOpd->pistonOPD, 
                    srcOpd->nbFrames * srcOpd->nbBases * sizeof(float));
            memcpy (dstOpd[band].sigmaPiston, srcOpd->sigmaPiston, 
                    srcOpd->nbFrames * srcOpd->nbBases * sizeof(float));
        }
    }
   
    return amdlibSUCCESS;
}

/** Useful macro to free all dynamically allocated structures */ 
#define amdlibMeanPiston_FREEALL()                       \
    amdlibFree2DArrayFloatWrapping(instantOpdPistonPtr); \
    amdlibFree2DArrayFloatWrapping(instantOpdSigmaPtr);  \
    amdlibFree2DArrayFloatWrapping(opdPistonPtr);        \
    amdlibFree2DArrayFloatWrapping(opdSigmaPtr);
/**
 * Compute mean piston. 
 *
 * @param instantOpd pointer to amdlibPISTON input data structure (bin of
 * intantaneous opds).
 * @param band current treated band.
 * @param iBin index where to store result in the output (binned) structure.
 * @param selectedFrames structure containing information relative to frame
 * selection. By default, all frames are selected.
 * @param opd pointer to amdlibPISTON output data structure (binned).
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibMeanPiston(amdlibPISTON    *instantOpd,
                                  amdlibBAND      band,
                                  int             iBin,
                                  amdlibSELECTION *selectedFrames,
                                  amdlibPISTON    *opd)
{
    int          iFrame; /* usually loop on number of frames */
    int          nbFrames = instantOpd->nbFrames;
    int          iBase; /*usually loop on nbBases*/
    int          nbBases = instantOpd->nbBases;
    static float pistonOPD[amdlibNBASELINE], sigma2[amdlibNBASELINE];
    float        **instantOpdPistonPtr = NULL;
    float        **opdPistonPtr = NULL;
    float        **instantOpdSigmaPtr = NULL;
    float        **opdSigmaPtr = NULL;
    static amdlibERROR_MSG errMsg;
    
    amdlibLogTrace("amdlibMeanPiston()");

    if (instantOpd->bandFlag[band] == amdlibFALSE)
    {
        amdlibLogError("Piston for band '%d' not ever computed", band);
        amdlibMeanPiston_FREEALL();
        return amdlibFAILURE;
    }
    
    opd->bandFlag[band] = instantOpd->bandFlag[band];

    instantOpdPistonPtr = 
        amdlibWrap2DArrayFloat(instantOpd->pistonOPDArray[band],
                               instantOpd->nbBases,
                               instantOpd->nbFrames, errMsg);
    if (instantOpdPistonPtr == NULL)
    {
        amdlibLogError("amdlibWrap2DArrayFloat() failed !");
        amdlibLogErrorDetail(errMsg);
        amdlibMeanPiston_FREEALL();
        return amdlibFAILURE;
    }
    instantOpdSigmaPtr = 
        amdlibWrap2DArrayFloat(instantOpd->sigmaPistonArray[band],
                               instantOpd->nbBases,
                               instantOpd->nbFrames, errMsg);
    if (instantOpdSigmaPtr == NULL)
    {
        amdlibLogError("amdlibWrap2DArrayFloat() failed !");
        amdlibLogErrorDetail(errMsg);
        amdlibMeanPiston_FREEALL();
        return amdlibFAILURE;
    }
    opdPistonPtr = amdlibWrap2DArrayFloat(opd->pistonOPDArray[band], 
                                          opd->nbBases, 
                                          opd->nbFrames, errMsg);
    if (opdPistonPtr == NULL)
    {
        amdlibLogError("amdlibWrap2DArrayFloat() failed !");
        amdlibLogErrorDetail(errMsg);
        amdlibMeanPiston_FREEALL();
        return amdlibFAILURE;
    }
    opdSigmaPtr = amdlibWrap2DArrayFloat(opd->sigmaPistonArray[band], 
                                         opd->nbBases, 
                                         opd->nbFrames, errMsg);
    if (opdSigmaPtr == NULL)
    {
        amdlibLogError("amdlibWrap2DArrayFloat() failed !");
        amdlibLogErrorDetail(errMsg);
        amdlibMeanPiston_FREEALL();
        return amdlibFAILURE;
    }

    /* if many measures, sigma2 is the rms of piston values (jitter) */
    if (nbFrames > 1)
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            if (selectedFrames->band[band].nbSelectedFrames[iBase] != 0)
            {
                /* compute mean piston */
                pistonOPD[iBase] = 0.0; 
                for (iFrame = 0; iFrame < nbFrames; iFrame++)
                {
                    if (selectedFrames->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                    {
                        pistonOPD[iBase] += instantOpdPistonPtr[iFrame][iBase];
                    }
                }
                opdPistonPtr[iBin][iBase] = pistonOPD[iBase] / 
                    selectedFrames->band[band].nbSelectedFrames[iBase];
                /* compute statistical rms on pistons */
                sigma2[iBase] = 0.0;
                for (iFrame = 0; iFrame < nbFrames; iFrame++)
                {
                    if (selectedFrames->band[band].isSelectedPt[iBase][iFrame] == amdlibTRUE)
                    {
                        pistonOPD[iBase] = instantOpdPistonPtr[iFrame][iBase];
                        sigma2[iBase] += amdlibPow2(pistonOPD[iBase] - 
                                                    opdPistonPtr[iBin][iBase]);
                    }
                }
                sigma2[iBase] /= selectedFrames->band[band].nbSelectedFrames[iBase]; 
                opdSigmaPtr[iBin][iBase] = sqrt(sigma2[iBase]);
            }
            else
            {
                opdPistonPtr[iBin][iBase] = 0.0;
                opdSigmaPtr[iBin][iBase] = 0.0;
            }
        }
    }            
    else /* Do nothing more than copy, only one frame !. Also, avoids problems
          * with high values of instantOpdSigma if it is amdlibBLANKING_VALUE */
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            if (selectedFrames->band[band].nbSelectedFrames[iBase] != 0)
            {
                opdPistonPtr[iBin][iBase] = instantOpdPistonPtr[0][iBase];
                opdSigmaPtr[iBin][iBase] = instantOpdSigmaPtr[0][iBase];
            }
        }
    }

    memcpy(opd->pistonOPD, opd->pistonOPDArray[band], 
           opd->nbBases * opd->nbFrames * sizeof(float));
    memcpy(opd->sigmaPiston, opd->sigmaPistonArray[band], 
           opd->nbBases * opd->nbFrames * sizeof(float));
    amdlibMeanPiston_FREEALL();
    return amdlibSUCCESS;
}
#undef amdlibMeanPiston_FREEALL


/** Useful macro to free all dynamically allocated structures */ 
#define amdlibComputePiston_FREEALL() free(x12); free(x23); free(x31);  \
    free(sx12); free(sx23); free(sx31);  \
    amdlibFree2DArrayWrapping((void **)tablePtr)
 
/**
 * Compute piston. 
 *
 * This function computes the piston_OPD for each frame and each baseline 
 * starting from the interspectra. Computes relevant seuance of data, and
 * passes them to amdlibComputePiston2T or amdlibComputePiston3T according
 * to the number of baselines.
 * 
 * @param instantCorrFlux instantaneous correlated flux.
 * @param band current treated band.
 * @param wave wavelength data structure used.
 * @param algoType indicates wether the piston is to be measured by fitting
 * a slope in the phasors (amdlibITERATIVE_PHASOR) or in the differential
 * phases after dewrapping (amdlibUNWRAPPED_PHASE).
 * @param instantOpd resulting computed piston.
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibComputePiston(amdlibVIS              *instantCorrFlux,
                                     /* instantCorrFlux because it is a subset 
                                      * of the orginal vis */
                                     amdlibBAND             band,
                                     amdlibWAVELENGTH       *wave,
                                     amdlibPISTON_ALGORITHM algoType,
                                     amdlibPISTON           *instantOpd,
                                     amdlibERROR_MSG        errMsg)
{
    int iFrame; /* usually loop on number of frames */
    int lVis;  /* the corresponding lambda index in the visibilities table */
    int nbBases = instantCorrFlux->nbBases;
    int nbFrames = instantCorrFlux->nbFrames;
    int nbLVis = instantCorrFlux->nbWlen;
    int nbTel = 2;
    amdlibCOMPLEX *x12 = NULL;  /* interspectrum base 12*/  /* Method 2 */
    amdlibCOMPLEX *x23 = NULL;  /* etc... */
    amdlibCOMPLEX *x31 = NULL;
    amdlibCOMPLEX *sx12 = NULL; /* corresponding errors */
    amdlibCOMPLEX *sx23 = NULL;
    amdlibCOMPLEX *sx31 = NULL;
    amdlibVIS_TABLE_ENTRY **tablePtr = NULL;

    x12 = calloc(instantCorrFlux->nbWlen, sizeof(amdlibCOMPLEX));
    if (x12 == NULL)
    {
        amdlibComputePiston_FREEALL();
        amdlibSetErrMsg("Allocation of x12 failed");
        return amdlibFAILURE;
    }
    x23 = calloc(instantCorrFlux->nbWlen, sizeof(amdlibCOMPLEX));
    if (x23 == NULL)
    {
        amdlibComputePiston_FREEALL();
        amdlibSetErrMsg("Allocation of x23 failed");
        return amdlibFAILURE;
    }
    x31 = calloc(instantCorrFlux->nbWlen, sizeof(amdlibCOMPLEX));
    if (x31 == NULL)
    {
        amdlibComputePiston_FREEALL();
        amdlibSetErrMsg("Allocation of x31 failed");
        return amdlibFAILURE;
    }
    sx12 = calloc(instantCorrFlux->nbWlen, sizeof(amdlibCOMPLEX));
    if (sx12 == NULL)
    {
        amdlibComputePiston_FREEALL();
        amdlibSetErrMsg("Allocation of sx12 failed");
        return amdlibFAILURE;
    }
    sx23 = calloc(instantCorrFlux->nbWlen, sizeof(amdlibCOMPLEX));
    if (sx23 == NULL)
    {
        amdlibComputePiston_FREEALL();
        amdlibSetErrMsg("Allocation of sx23 failed");
        return amdlibFAILURE;
    }
    sx31 = calloc(instantCorrFlux->nbWlen, sizeof(amdlibCOMPLEX));
    if (sx31 == NULL)
    {
        amdlibComputePiston_FREEALL();
        amdlibSetErrMsg("Allocation of sx31 failed");
        return amdlibFAILURE;
    }
    

    /* Wrap instantCorrFlux->table structure to easy the reading of the code*/
    tablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(instantCorrFlux->table,
                                                nbBases, nbFrames,
                                                sizeof(amdlibVIS_TABLE_ENTRY),
                                                errMsg);
    if (tablePtr == NULL)
    {
        amdlibComputePiston_FREEALL();
        return amdlibFAILURE;
    }
    
    if(nbBases == 3)
    {
        nbTel = 3;
    }
    for (iFrame=0; iFrame < nbFrames; iFrame++)
    {
        for (lVis=0; lVis < nbLVis; lVis++)
        {
            /* baseline 1-2 */

            x12[lVis].re = tablePtr[iFrame][0].vis[lVis].re;
            x12[lVis].im = tablePtr[iFrame][0].vis[lVis].im;
            sx12[lVis].re = tablePtr[iFrame][0].sigma2Vis[lVis].re;
            sx12[lVis].im = tablePtr[iFrame][0].sigma2Vis[lVis].im;
            
            if (nbTel==3)
            {
                /* baseline 2-3 */
                x23[lVis].re = tablePtr[iFrame][1].vis[lVis].re;
                x23[lVis].im = tablePtr[iFrame][1].vis[lVis].im;
                sx23[lVis].re = tablePtr[iFrame][1].sigma2Vis[lVis].re;
                sx23[lVis].im = tablePtr[iFrame][1].sigma2Vis[lVis].im;

                /* baseline 3-1 */
                x31[lVis].re = tablePtr[iFrame][2].vis[lVis].re;
                x31[lVis].im = tablePtr[iFrame][2].vis[lVis].im;    
                sx31[lVis].re = tablePtr[iFrame][2].sigma2Vis[lVis].re;
                sx31[lVis].im = tablePtr[iFrame][2].sigma2Vis[lVis].im;    
            }
        }


        if (nbTel==3)
        {
            amdlibComputePiston3T(nbLVis, wave->wlen, x12, x23, x31,
                        sx12, sx23, sx31,
                        &instantOpd->pistonOPDArray[band][nbBases*iFrame],
                        &instantOpd->sigmaPistonArray[band][nbBases*iFrame],
                        algoType);
        }
        else
        {
            amdlibComputePiston2T(nbLVis, wave->wlen, x12, sx12,
                        &instantOpd->pistonOPDArray[band][nbBases*iFrame],
                        &instantOpd->sigmaPistonArray[band][nbBases*iFrame],
                        algoType);
        }
    }  /* end loop iFrame */

    instantOpd->bandFlag[band] = amdlibTRUE;
    memcpy(instantOpd->pistonOPD, instantOpd->pistonOPDArray[band],
           instantOpd->nbFrames * instantOpd->nbBases * sizeof(float));
    memcpy(instantOpd->sigmaPiston, instantOpd->sigmaPistonArray[band],
           instantOpd->nbFrames * instantOpd->nbBases * sizeof(float));
    amdlibComputePiston_FREEALL();
    return amdlibSUCCESS;
}
#undef amdlibComputePiston_FREEALL

#define amdlibComputePistonIterative2T_FREEALL() free(waveNr); free(totoPist); \
    free(cVisC); free(cVisCN); free(cVisC0); free(cVisCN0); free(cVisC1);      \
    free(cVisCN1); free(cVisC2); free(cVisCN2); free(chi2); free(cVisCTable);  \
    free(cVisCNTable);

/**
 * Compute piston, iterative phasor method. Normally shoudl be the only
 * one method used, but for historical reasons, the unwrapped_phase was
 * used by the instrument and we have to keep its wrong sign, and maintain
 * the unwrapped_phase methos as well.
 */
void amdlibComputePistonIterative2T(int                    nbLVis,
                                    float                  *wlen,
                                    amdlibCOMPLEX          *cpxVisTable,
                                    amdlibCOMPLEX          *sigma2CpxVisTable,
                                    float                  *pistonOPD,
                                    float                  *sigma)
{
    int lVis;
    int i, iMin, N;
    amdlibCOMPLEX phasor;
    amdlibERROR_MSG errMsg;
    
    double *waveNr, *totoPist, *chi2;
    amdlibCOMPLEX *cVisC, *cVisCN, *cVisC0, *cVisCN0,*cVisC1, *cVisCN1, 
                  *cVisC2, *cVisCN2, avgVis, rmsVis, conjAvgVis;
    amdlibCOMPLEX **cVisCTable, **cVisCNTable;

    double x, R=0, wlenAvg=0, wlenDifAvg=0;
    double opd0, opd1, opd2, Dopd, chi20, chi21, chi22;
    double r;
    double coefGrow = 1.5, coefShrink = 2.;
    
    waveNr = calloc(nbLVis, sizeof(*waveNr)); 
    chi2 = calloc(nbLVis, sizeof(*chi2)); 
    totoPist = calloc(nbLVis, sizeof(*totoPist)); 
    cVisC = calloc(nbLVis*nbLVis, sizeof(*cVisC)); 
    cVisCN = calloc(nbLVis*nbLVis, sizeof(*cVisCN)); 
    cVisC0 = calloc(nbLVis, sizeof(*cVisC0)); 
    cVisCN0 = calloc(nbLVis, sizeof(*cVisCN0)); 
    cVisC1 = calloc(nbLVis, sizeof(*cVisC1)); 
    cVisCN1 = calloc(nbLVis, sizeof(*cVisCN1)); 
    cVisC2 = calloc(nbLVis, sizeof(*cVisC2)); 
    cVisCN2 = calloc(nbLVis, sizeof(*cVisCN2)); 
    cVisCTable = (amdlibCOMPLEX **)amdlibWrap2DArray(cVisC, nbLVis, nbLVis,
                                        sizeof(amdlibCOMPLEX),
                                        errMsg);
    cVisCNTable = (amdlibCOMPLEX **)amdlibWrap2DArray(cVisCN,nbLVis, nbLVis,
                                        sizeof(amdlibCOMPLEX),
                                        errMsg);
    
    for (lVis = 0; lVis < nbLVis; lVis++)
    {
        waveNr[lVis] = 2 * M_PI / wlen[lVis];
        wlenAvg += wlen[lVis];
    }
    wlenAvg /= nbLVis;
    for (lVis = 1; lVis < nbLVis; lVis++)
    {
        wlenDifAvg += (wlen[lVis] - wlen[lVis-1]);
    }
    wlenDifAvg /= (nbLVis - 1);
    R = wlenAvg / (2 * wlenDifAvg);
    for (lVis = 0; lVis < nbLVis; lVis++)
    {
        totoPist[lVis] = -R * wlenAvg / 2 + lVis * R * wlenAvg / nbLVis - 1;
    }

    for (lVis = 0; lVis < nbLVis; lVis++)
    {
        for (i = 0; i < nbLVis; i++)
        {
            x = waveNr[i] * totoPist[lVis];
            phasor.re = cos(x);
            phasor.im = -sin(x);
            amdlibCpxMul(cpxVisTable[i], phasor, cVisCTable[lVis][i]);
        }
    }

    for (lVis = 0; lVis < nbLVis; lVis++)
    {
        conjAvgVis.re = 0.0;
        conjAvgVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            conjAvgVis.re += cVisCTable[lVis][i].re;
            conjAvgVis.im -= cVisCTable[lVis][i].im;
        }
        conjAvgVis.re /= nbLVis ;
        conjAvgVis.im /= nbLVis ;
        for (i = 0; i < nbLVis; i++)
        {
            amdlibCpxMul(cVisCTable[lVis][i], conjAvgVis, cVisCNTable[lVis][i]);
        }
    }
    /* rms cVisCN */
    for (lVis = 0; lVis < nbLVis; lVis++)
    {
        avgVis.re = 0.0;
        avgVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            avgVis.re += cVisCNTable[lVis][i].re;
            avgVis.im += cVisCNTable[lVis][i].im;
        }
        avgVis.re /= nbLVis ;
        avgVis.im /= nbLVis ;
        rmsVis.re = 0.0;
        rmsVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            rmsVis.re += amdlibPow2(cVisCNTable[lVis][i].re-avgVis.re);
            rmsVis.im += amdlibPow2(cVisCNTable[lVis][i].im-avgVis.im);
        }
        rmsVis.re /= nbLVis ;
        rmsVis.im /= nbLVis ;
        rmsVis.re = sqrt(rmsVis.re);
        rmsVis.im = sqrt(rmsVis.im);
        chi2[lVis] = rmsVis.re * rmsVis.im / avgVis.re;
    }
    iMin = amdlibFindIndexOfMinimum(chi2, nbLVis);
    srand((unsigned int)iMin);
    opd0 = totoPist[iMin];
    
    Dopd = 0.0;
    
    for (N = 0; N < amdlibMAX_PISTON2_ITERATIONS; N++)
    {
        for (lVis = 0; lVis < nbLVis; lVis++)
        {
            x = waveNr[lVis] * opd0;
            phasor.re = cos(x);
            phasor.im = -sin(x);
            amdlibCpxMul(cpxVisTable[lVis], phasor, cVisC0[lVis]);
        }
        conjAvgVis.re = 0.0;
        conjAvgVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            conjAvgVis.re += cVisC0[i].re;
            conjAvgVis.im -= cVisC0[i].im;
        }
        conjAvgVis.re /= nbLVis ;
        conjAvgVis.im /= nbLVis ;
        for (lVis = 0; lVis < nbLVis; lVis++)
        {
            amdlibCpxMul(cVisC0[lVis], conjAvgVis, cVisCN0[lVis]);
        }
        avgVis.re = 0.0;
        avgVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            avgVis.re += cVisCN0[i].re;
            avgVis.im += cVisCN0[i].im;
        }
        avgVis.re /= nbLVis ;
        avgVis.im /= nbLVis ;
        rmsVis.re = 0.0;
        rmsVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            rmsVis.re += amdlibPow2(cVisCN0[i].re - avgVis.re);
            rmsVis.im += amdlibPow2(cVisCN0[i].im - avgVis.im);
        }
        rmsVis.re /= nbLVis ;
        rmsVis.im /= nbLVis ;
        rmsVis.re = sqrt(rmsVis.re);
        rmsVis.im = sqrt(rmsVis.im);
        chi20 = rmsVis.re * rmsVis.im / avgVis.re;
        
        opd1 = opd0 + Dopd;
        for (lVis = 0; lVis < nbLVis; lVis++)
        {
            x = waveNr[lVis] * opd1;
            phasor.re = cos(x);
            phasor.im = -sin(x);
            amdlibCpxMul(cpxVisTable[lVis], phasor, cVisC1[lVis]);
        }
        conjAvgVis.re = 0.0;
        conjAvgVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            conjAvgVis.re += cVisC1[i].re;
            conjAvgVis.im -= cVisC1[i].im;
        }
        conjAvgVis.re /= nbLVis ;
        conjAvgVis.im /= nbLVis ;
        for (lVis = 0; lVis < nbLVis; lVis++)
        {
            amdlibCpxMul(cVisC1[lVis], conjAvgVis, cVisCN1[lVis]);
        }
        avgVis.re = 0.0;
        avgVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            avgVis.re += cVisCN1[i].re;
            avgVis.im += cVisCN1[i].im;
        }
        avgVis.re /= nbLVis ;
        avgVis.im /= nbLVis ;
        rmsVis.re = 0.0;
        rmsVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            rmsVis.re += amdlibPow2(cVisCN1[i].re - avgVis.re);
            rmsVis.im += amdlibPow2(cVisCN1[i].im - avgVis.im);
        }
        rmsVis.re /= nbLVis ;
        rmsVis.im /= nbLVis ;
        rmsVis.re = sqrt(rmsVis.re);
        rmsVis.im = sqrt(rmsVis.im);
        chi21 = rmsVis.re * rmsVis.im / avgVis.re;
        
        opd2 = opd0 - Dopd;
        for (lVis = 0; lVis < nbLVis; lVis++)
        {
            x = waveNr[lVis] * opd2;
            phasor.re = cos(x);
            phasor.im = -sin(x);
            amdlibCpxMul(cpxVisTable[lVis], phasor, cVisC2[lVis]);
        }
        conjAvgVis.re = 0.0;
        conjAvgVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            conjAvgVis.re += cVisC2[i].re;
            conjAvgVis.im -= cVisC2[i].im;
        }
        conjAvgVis.re /= nbLVis ;
        conjAvgVis.im /= nbLVis ;
        for (lVis = 0; lVis < nbLVis; lVis++)
        {
            amdlibCpxMul(cVisC2[lVis], conjAvgVis, cVisCN2[lVis]);
        }
        avgVis.re = 0.0;
        avgVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            avgVis.re += cVisCN2[i].re;
            avgVis.im += cVisCN2[i].im;
        }
        avgVis.re /= nbLVis ;
        avgVis.im /= nbLVis ;
        rmsVis.re = 0.0;
        rmsVis.im = 0.0;
        for (i = 0; i < nbLVis; i++)
        {
            rmsVis.re += amdlibPow2(cVisCN2[i].re - avgVis.re);
            rmsVis.im += amdlibPow2(cVisCN2[i].im - avgVis.im);
        }
        rmsVis.re /= nbLVis ;
        rmsVis.im /= nbLVis ;
        rmsVis.re = sqrt(rmsVis.re);
        rmsVis.im = sqrt(rmsVis.im);
        chi22 = rmsVis.re * rmsVis.im / avgVis.re;

        if ((chi21 >= chi20) && (chi20 > chi22))
        {
            opd0 = opd0 - Dopd / coefShrink;
            Dopd = Dopd * coefGrow;
        }
        else if ((chi21 <= chi20) && (chi20 < chi22))
        {
            opd0 = opd0 + Dopd / coefShrink;
            Dopd = Dopd * coefGrow;
        }
        else if ((chi21 > chi20) && (chi20 < chi22))
        {
            Dopd = Dopd / coefShrink;
        }
        else
        {
            r = (double)rand() / (double)RAND_MAX;
            if (Dopd == 0)
            {
                Dopd = (r - 0.5) * wlen[0];
            }
            if (r > 0.5)
            {
                opd0 -= Dopd / coefShrink;
            }
            else
            {
                opd0 += Dopd / coefShrink;
            }
        }
        if  (amdlibPow2(Dopd / opd0) < amdlibPow2(amdlibPRECISION))
        { 
            break;
        }
    }
    /* here we multiply by -1 for same sign as the other method.*/
    *pistonOPD =  -1*opd0;
    *sigma = sqrt(fabs(opd0 * chi20));

    amdlibComputePistonIterative2T_FREEALL();
}

/**
 * Display piston structure.
 *
 * @param opd pointer to amdlibPISTON structure.
 */
void amdlibDisplayPiston(amdlibPISTON *opd)
{
    int nbFrames, nbBases;
    int iBand, iFrame, iBase, iCell;
    amdlibBOOLEAN  bandFlag;
    float bandPistonOPD;
    float bandSigmaPiston;

    amdlibLogTrace("amdlibDisplayPiston()");

    /* Display number of frames */
    nbFrames = opd->nbFrames;
    printf("nbFrames = %d\n", nbFrames);
    
    /* Display number of base */
    nbBases = opd->nbBases;
    printf("nbBases = %d\n", nbBases);

    /* Display arrays[amdlibNB_BAND] */
    for (iBand = 0; iBand < amdlibNB_BANDS; iBand++)
    {
        /* bandFlag */
        bandFlag = opd->bandFlag[iBand];
        printf("bandFlag[%d] = %d\n", iBand, bandFlag);
    }
    /* Display different arrays[nbBand][nbFrames][nbBases] */
    for (iBand = 0; iBand < amdlibNB_BANDS; iBand++)
    {
        for (iFrame = 0; iFrame < nbFrames; iFrame++)
        {
            for (iBase = 0; iBase < nbBases; iBase++)
            {            
                /* Set cell[iFrame][iBase] */
                iCell = iFrame * nbBases + iBase;
                printf("---> band[%d] - cell frame/base[%d][%d]\n",iBand, 
                       iFrame, iBase);
                
                /* pistonOPDArray */
                bandPistonOPD = opd->pistonOPDArray[iBand][iCell];
                printf("pistonOPDArray[%d][%d][%d] = %f - ", iBand, iFrame, 
                       iBase, bandPistonOPD);

                /* sigmaPistonArray */
                bandSigmaPiston = opd->sigmaPistonArray[iBand][iCell];
                printf("sigmaPistonArray[%d][%d][%d] = %f\n", iBand, iFrame,
                       iBase, bandSigmaPiston);
            }
        }
    }
}

/*
 * Local functions
 */
/** Useful macro to free all dynamically allocated structures */ 
#define amdlibComputePiston2T_FREEALL() free(sigma2_dPhi); free(phiRes); \
    free(pst); free(dSigma); free(newXim); free(newXre); free(isIm);     \
    free(isRe); free(s);

/**
 * Compute piston - method for 2 telescopes. 
 *
 * This function computes the piston_OPD for each frame and each baseline 
 * starting from the interspectra, method for 2T 
 * 
 * @param nbWlen number of wavelengths.
 * @param Wave array of wavelengths.
 * @param x array containing the correlated fluxes expressed as complexes.
 * @param sx array containing the variance of x.
 * @param pistonOPD pointer to the computed piston value to be returned.
 * @param sigma pointer to the error on pistonOPD to be returned.
 * @param algoType indicates wether the piston is to be measured by fitting
 * a slope in the phasors (amdlibITERATIVE_PHASOR) or in the differential
 * phases after dewrapping (amdlibUNWRAPPED_PHASE).
 */
void amdlibComputePiston2T(int                    nbWlen,
                           float                  *Wave,
                           amdlibCOMPLEX          *x,
                           amdlibCOMPLEX          *sx,
                           float                  *pistonOPD,
                           float                  *sigma,
                           amdlibPISTON_ALGORITHM algoType)
{
    int    l;
    double avgIsRe, avgIsIm, avgDSigma;
    double *s, *isRe, *isIm;
    double avgXre,avgXim;
    double *newXre, *newXim;
    double *dSigma, *pst;
    double *phiRes;
    double *sigma2_dPhi;
    double dPhi;
    double N1, N2, amp_is;
    double rms;


    amdlibLogTrace("amdlibComputePiston2T()");
    
    if (algoType == amdlibITERATIVE_PHASOR)
    {
        amdlibComputePistonIterative2T
        (nbWlen, Wave, x, sx, pistonOPD, sigma);
        /* Note Piston Sign will be reversed in amdlibComputePistonIterative2T
         * to avoid conflicts with the (wrong sign) piston returned by the
         * UNWRAPPED_PHASE method used by the instrument*/
    }
    else if (algoType == amdlibUNWRAPPED_PHASE)
    {
        /* Note Piston Sign returned by this method is believed to be wrong
         * but is kept as is for historical reasons */

        s = calloc(nbWlen-1, sizeof(double)); 
        isRe = calloc(nbWlen-1, sizeof(double));
        isIm = calloc(nbWlen-1, sizeof(double));
        newXre = calloc(nbWlen, sizeof(double));
        newXim = calloc(nbWlen, sizeof(double));
        dSigma = calloc(nbWlen-1, sizeof(double));
        pst = calloc(nbWlen-1, sizeof(double));
        phiRes = calloc(nbWlen-1, sizeof(double));
        sigma2_dPhi = calloc(nbWlen-1, sizeof(double));

        /* First estimation of the piston */
        pistonOPD[0] = 0.0;
        for (l = 0; l < nbWlen - 1; l++)
        {
            /* First calculate the interspectra of the coherent flux */
            isRe[l]  = (x[l].re * x[l+1].re + x[l].im * x[l+1].im);
            isIm[l]  = (x[l].im * x[l+1].re - x[l].re * x[l+1].im);

            /* Calculate the number of photons used to estimate phase error */
            N1 = sqrt(x[l+1].re * x[l+1].re + x[l+1].im * x[l+1].im);
            N2 = sqrt(x[l].re * x[l].re + x[l].im * x[l].im);

            amp_is = sqrt(isRe[l] * isRe[l] + isIm[l] * isIm[l]);

            /* Calculate the phase error */
            sigma2_dPhi[l] =  (1/2. * (N1 * N2 + (N1 + N2) * amp_is)) /
                amdlibPow2(amp_is);

            if(sigma2_dPhi[l] > 1)
                sigma2_dPhi[l] = amdlibBLANKING_VALUE;

            /* Calculate the differential of phase */
            dPhi = atan2(isIm[l], isRe[l]);

            /* Calculate the differential of wave number */
            dSigma[l] = 1.0 / Wave[l+1] - 1.0 / Wave[l];

            pst[l] = (dPhi / (2 * M_PI * dSigma[l]));
        }

        pistonOPD[0] = amdlibAvgTable(nbWlen-1, pst, sigma2_dPhi);

        /* Calculate the residual phase */
        for (l = 0; l < nbWlen; l++)
        {
            newXre[l] = (x[l].re * cos(2 * (M_PI / Wave[l]) * pistonOPD[0]) -
                         x[l].im * sin(2 * (M_PI / Wave[l]) * pistonOPD[0]));
            newXim[l] = (x[l].re * sin(2 * (M_PI / Wave[l]) * pistonOPD[0]) +
                         x[l].im * cos(2 * (M_PI / Wave[l]) * pistonOPD[0]));
        }

        avgXre = amdlibAvgTable(nbWlen-1, newXre, sigma2_dPhi);
        avgXim = amdlibAvgTable(nbWlen-1, newXim, sigma2_dPhi);

        dPhi = atan2(avgXim, avgXre);

        /* Correct the piston estimation with this residual */
        pistonOPD[0] += dPhi / (4*M_PI*(1.0 / Wave[nbWlen-1] - 1.0 / Wave[0]));


        /* Calculate the residual rms */
        for (l = 0; l < nbWlen - 1; l++)
        {
            newXre[l] += (x[l].re * cos(2 * (M_PI / Wave[l]) * pistonOPD[0]) -
                          x[l].im * sin(2 * (M_PI / Wave[l]) * pistonOPD[0]));
            newXim[l] += (x[l].re * sin(2 * (M_PI / Wave[l]) * pistonOPD[0]) +
                          x[l].im * cos(2 * (M_PI / Wave[l]) * pistonOPD[0]));
            phiRes[l] = atan2(newXim[l], newXre[l]);
        }

        rms = amdlibRmsTable(nbWlen-1, phiRes, sigma2_dPhi);

        /* If the residual phase rms is more that 1 radian, then use
         * the second method to compute the phase */      
        if(rms > 1.0)
        {
            for (l = 0; l < nbWlen - 1; l++)
            {
                /* calculate again the interspectra of the redressed
                   coherent flux */
                isRe[l]  = (newXre[l] * newXre[l+1] + newXim[l] * newXim[l+1]);
                isIm[l]  = (newXim[l] * newXre[l+1] - newXre[l] * newXim[l+1]);
            }
            avgIsRe = amdlibAvgTable(nbWlen-1, isRe, sigma2_dPhi);
            avgIsIm = amdlibAvgTable(nbWlen-1, isIm, sigma2_dPhi);
            dPhi = atan2(avgIsIm, avgIsRe);

            avgDSigma = amdlibAvgTable(nbWlen-1, dSigma, sigma2_dPhi);

            pistonOPD[0] += dPhi / (2 * M_PI * avgDSigma);


            /* Calculate the residual phase again */
            for (l = 0; l < nbWlen - 1; l++)
            {
                newXre[l] = (x[l].re * cos(2*(M_PI / Wave[l]) * pistonOPD[0]) -
                             x[l].im * sin(2*(M_PI / Wave[l]) * pistonOPD[0]));
                newXim[l] = (x[l].re * sin(2*(M_PI / Wave[l]) * pistonOPD[0]) +
                             x[l].im * cos(2*(M_PI / Wave[l]) * pistonOPD[0]));
            }

            avgXre = amdlibAvgTable(nbWlen-1, newXre, sigma2_dPhi);
            avgXim = amdlibAvgTable(nbWlen-1, newXim, sigma2_dPhi);

            dPhi = atan2(avgXim, avgXre);   

            /* Correct the piston estimation with this residual */
	  
            pistonOPD[0] += dPhi / (4*M_PI*(1.0 / Wave[nbWlen-1] - 
                                            1.0 / Wave[0]));

            /* Calculate the residual rms */
            for (l = 0; l < nbWlen - 1; l++)
            {
                newXre[l] += (x[l].re * cos(2*(M_PI / Wave[l]) * pistonOPD[0]) -
                              x[l].im * sin(2*(M_PI / Wave[l]) * pistonOPD[0]));
                newXim[l] += (x[l].re * sin(2*(M_PI / Wave[l]) * pistonOPD[0]) +
                              x[l].im * cos(2*(M_PI / Wave[l]) * pistonOPD[0]));
                phiRes[l] = atan2(newXim[l], newXre[l]);
            }

            rms = amdlibRmsTable(nbWlen-1, phiRes, sigma2_dPhi);
        }
      
        /* Calculate the error on the piston */
        if(rms < 1)
        {
            *sigma = rms / (4*M_PI*(1.0 / Wave[nbWlen-1] - 1.0 / Wave[0]));
        }
        else 
        {
            *sigma = amdlibBLANKING_VALUE;
        }

#if amdlibDEBUG
        amdlibLogTrace("Debugging amdlibComputePiston2T\n");

        amdlibLogTrace("nbWlen : %d",nbWlen);
        amdlibLogTrace("Piston: %f +- %f", pistonOPD[0],*sigma);
        amdlibLogTrace("Wlen end: %f Wlen start: %f",Wave[nbWlen-1],Wave[0]);

        amdlibLogTrace("Wlen :[");      
        for(l=0; l<nbWlen-1; l++)
        {
            amdlibLogTrace("%f, ",Wave[l]);
        }
        amdlibLogTrace("%f]",Wave[nbWlen-1]);

        amdlibLogTrace("Phi :[");      
        for(l=0; l<nbWlen-1; l++)
        {
            amdlibLogTrace("%f, ",atan2(x[l].im,x[l].re));
        }
        amdlibLogTrace("%f]",atan2(x[nbWlen-1].im,x[nbWlen-1].re));

#endif
    amdlibComputePiston2T_FREEALL();
    }
}

#undef amdlibComputePiston2T_FREEALL


/** Useful macro to free all dynamically allocated structures */ 
#define amdlibComputePiston3T_FREEALL() \
    free(s12);free(s23);free(s31);

/**
 * Compute piston - method for 3 telescopes. 
 *
 * This function computes the piston_OPD for each frame and each baseline 
 * starting from the interspectra, method for 3T 
 * 
 * @param nbWlen number of wavelengths.
 * @param Wave array of wavelengths.
 * @param x12 array containing the visibility expressed as a complex for
 * baseline 1-2.
 * @param x23 array containing the visibility expressed as a complex for
 * baseline 2-3.
 * @param x31 array containing the visibility expressed as a complex for
 * baseline 3-1.
 * @param sx12 array containing the variances of x12
 * @param sx23 array containing the variances of x23
 * @param sx31 array containing the variances of x31
 * @param pistonOPD pointer to the computed piston values to be returned.
 * @param sigma pointer to the errors on pistonOPD to be returned.
 * @param algoType indicates wether the piston is to be measured by fitting
 * a slope in the phasors (amdlibITERATIVE_PHASOR) or in the differential
 * phases after dewrapping (amdlibUNWRAPPED_PHASE).
 */
void amdlibComputePiston3T(int                    nbWlen,
                           float                  *Wave,
                           amdlibCOMPLEX          *x12,
                           amdlibCOMPLEX          *x23, 
                           amdlibCOMPLEX          *x31,
                           amdlibCOMPLEX          *sx12,
                           amdlibCOMPLEX          *sx23, 
                           amdlibCOMPLEX          *sx31,
                           float                  *pistonOPD,
                           float                  *sigma,
                           amdlibPISTON_ALGORITHM algoType)
{


    amdlibLogTrace("amdlibComputePiston3T()");
    
    if(algoType == amdlibITERATIVE_PHASOR)
    {
        /* Note Piston Sign will be reversed in amdlibComputePistonIterative2T
         * to avoid conflicts with the (wrong sign) piston returned by the
         * UNWRAPPED_PHASE method used by the instrument*/
        amdlibComputePistonIterative2T
        (nbWlen, Wave, x12, sx12, 
         &pistonOPD[0], &sigma[0]);
        amdlibComputePistonIterative2T
        (nbWlen, Wave, x23, sx23, 
         &pistonOPD[1], &sigma[1]);
        amdlibComputePistonIterative2T
        (nbWlen, Wave, x31, sx31, 
         &pistonOPD[2], &sigma[2]);
        
    }
    else if(algoType == amdlibUNWRAPPED_PHASE)
    {
        /* Note Piston Sign returned by this method is believed to be wrong
         * but is kept as is for historical reasons */
        amdlibComputePiston2T( nbWlen, Wave, x12, sx12,
                               &pistonOPD[0], &sigma[0],
                               amdlibUNWRAPPED_PHASE);

#if amdlibDEBUG
        amdlibLogTrace("OPD 12 : %f        err12 : %f",pistonOPD[0],sigma[0]);
#endif
        amdlibComputePiston2T(nbWlen, Wave, x23, sx23,
                              &pistonOPD[1], &sigma[1],
                              amdlibUNWRAPPED_PHASE);

#if amdlibDEBUG
        amdlibLogTrace("OPD 23 : %f        err23 : %f",pistonOPD[1],sigma[1]);
#endif

        amdlibComputePiston2T(nbWlen, Wave, x31, sx31,
                              &pistonOPD[2], &sigma[2],
                              amdlibUNWRAPPED_PHASE);

#if amdlibDEBUG
        amdlibLogTrace("OPD 31 : %f        err31 : %f",pistonOPD[2],sigma[2]);      
#endif
    }
}  /* end amdlibComputePiston3T */

#undef amdlibComputePiston3T_FREEALL

/**
 * Free memory allocated for pistons. 
 *
 * @param piston pointer to amdlibPISTON structure.
 */
void amdlibFreePiston(amdlibPISTON *piston)
{
    int band;
    
    amdlibLogTrace("amdlibFreePiston()");
    
    /* Check thisPtr and do nothing if unitialized*/
    if (piston->thisPtr != piston)
    {
        return;
    }

    for (band=amdlibJ_BAND; band <= amdlibK_BAND; band++)
    {
        if (piston->pistonOPDArray[band] != NULL)
        {
            free(piston->pistonOPDArray[band]);
        }
        piston->pistonOPDArray[band] = NULL;
        if (piston->sigmaPistonArray[band] != NULL)
        {
            free(piston->sigmaPistonArray[band]);
        }
        piston->sigmaPistonArray[band] = NULL;
    }
    
    if (piston->pistonOPD != NULL)
    {
        free(piston->pistonOPD);
    }
    piston->pistonOPD = NULL;
    if (piston->sigmaPiston != NULL)
    {
        free(piston->sigmaPiston);
    }
    piston->sigmaPiston = NULL;

    piston->thisPtr = NULL;
}

/*___oOo___*/
