/************************************************************************
 * E.S.O. - VLT project
 *
 * who       when      what
 * --------  --------  ----------------------------------------------
 * gzins     12/02/03  Created
 * gzins     29/04/03  Added documentation
 * gzins     29/07/03  Added function for quick-look
 * gduvert   29/07/04  Fixed bug related to the computation of the detector
 *                     read-out noise
 * gduvert   26/10/04  - change documentation
 *                     - add remanence substracting
 * gduvert   15/12/04  image pointer was wrongly declared static. Removed.
 * mkiekebu 2004-12-19 added amdlibStripQuotes
 * gduvert   02/02/05  better use of calloc(), calloc() zeroes arrays, so
 *                     memset() is not needed. 
 * tlicha    05/04/05  changed amdlibCalibrateRawData memory allocation to
 *                     be compliant with GASGANO/CPL
 * gduvert   26/05/05  set expTime to dit
 * gduvert   25/08/05  changes to handle spurious NULL frames in RAW Data.
 *                     Compress in-place the raw data structure to forget
 *                     these frames during the amdlibCalibrateRawData phase.
 *                     Used a trick when writing back a calibrated raw data
 *                     file to FITS by recomputing the PCOUNT header value of
 *                     the table.
 * gzins     26/01/06  fixed bug related to the counting of saturated pixels,
 *                     and memory handling.
 * ealtariba 2006-02-24 no longer read NDIT; not used
 * ealtariba 2006-07-04 corrected minor bugs.

 */

/**
 * @file
 * Functions to handle raw data. 
 *
 * The raw data is the data produced by the instrument during an exposure. This
 * data is stored as binary tables in a FIT file. 
 */
#define _POSIX_SOURCE 1
#define amdlibREMANENCE 0.00

static char *rcsId="@(#) $Id: amdlibRawData.c,v 1.5 2011/01/24 12:43:55 agabasch Exp $";
static void *use_rcsId = (0 ? &use_rcsId : (void*)&rcsId);

/*
 * System Headers
 */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include "fitsio.h"
#include <time.h>

/*
 * Local Headers
 */
#include "amdlib.h"
#include "amdlibProtected.h"

/* List of keywords corresponding to instrument configuration */
static char *amdlibRequiredInsCfgKeywList[] =
{
    "HIERARCH ESO INS GRAT1 NAME",
    "HIERARCH ESO INS GRAT1 ORDER",
    "HIERARCH ESO INS GRAT1 WLEN",
    "HIERARCH ESO INS GRIS1 NAME",
    "HIERARCH ESO INS GRIS2 NAME",
    "HIERARCH ESO INS OPTI2 NAME",
    "HIERARCH ESO INS OPTI5 NAME",
    "HIERARCH ESO INS OPTI6 NAME",
    "HIERARCH ESO INS OPTI7 NAME",
    NULL
};

/*
 * Local functions
 */
static amdlibCOMPL_STAT amdlibReadTimeTag (fitsfile        *filePtr, 
                                           amdlibRAW_DATA  *rawData,
                                           int             firstFrame,
                                           int             nbFrames,
                                           amdlibERROR_MSG errMsg);
static amdlibCOMPL_STAT amdlibWriteTimeTag (fitsfile        *filePtr, 
                                            amdlibRAW_DATA  *rawData,
                                            amdlibERROR_MSG errMsg);
static float amdlibMedian9(float *data);
static void  amdlibFreeRawData(amdlibRAW_DATA *rawData);
static amdlibFRAME_TYPE amdlibFrameTypeStr2Id(char *frameTypeStr);
static int amdlibFindPeakPosition(double *data, int nbData, double threshold,
                                  int nbMaxPeaks, double *peakPos, 
                                  double *peakWeight, double *peakMax);
static amdlibCOMPL_STAT amdlibCheckDataCompatibility(
                                              amdlibPIXEL_BIAS_DATA *pixelBias,
                                              amdlibRAW_DATA        *rawData,
                                              amdlibERROR_MSG       errMsg);
static amdlibCOMPL_STAT amdlibDiscardNullFrames(amdlibRAW_DATA  *rawData,
                                                amdlibERROR_MSG errMsg);
static amdlibCOMPL_STAT amdlibSubtractPixelBias(amdlibPIXEL_BIAS_DATA *pixelBias,
                                            amdlibRAW_DATA        *rawData,
                                            amdlibERROR_MSG       errMsg);
static amdlibCOMPL_STAT amdlibCalculateRowBias(
                              amdlibRAW_DATA        *rawData,
                              int                   iFrame,
                              float                 gain,
                              int                   *nbGoodPixelsOnDetector, 
                              float                 *rowBias,
                              amdlibERROR_MSG       errMsg);
static amdlibCOMPL_STAT amdlibCalibrateImage(
                              amdlibRAW_DATA        *rawData,
                              int                   iFrame,
                              float                 gain,
                              float                 *rowBias,
                              int                   *nbSaturatedPixels,
                              amdlibERROR_MSG       errMsg);
static amdlibCOMPL_STAT amdlibSubtractRemanence(amdlibRAW_DATA  *rawData,
                                                amdlibERROR_MSG errMsg);
/*
 * Public functions
 */
/**
 * Load all frames from raw data file.
 *
 * This function loads the raw data file and updates all the raw data structures
 * accordingly. This concerns information related to:
 *      \li the instrument configuration
 *      \li the spectral channel data
 *      \li the imaging detector 
 *      \li the imaging data
 *      \li the region data
 *
 * @param filename name of the FITS file containing raw data.
 * @param rawData structure where loaded information will be stored
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibLoadRawData(const char      *filename,
                                   amdlibRAW_DATA  *rawData,
                                   amdlibERROR_MSG errMsg)
{
    amdlibLogTrace("amdlibLoadRawData()");
    
    return (amdlibLoadRawFrames(filename, rawData, amdlibFIRST_FRAME, 
                                amdlibALL_FRAMES, errMsg));
}

/**
 * Load bad pixel map from file. 
 *
 * This function loads, as amdlibLoadRawData() function does, the raw data file,
 * but it does not allocated memory to store region data and, of course, does
 * not them. This function is usefull to get information such the number of
 * frames, the number of region or the frame type without loading all the file.
 *
 * @param filename name of the FITS file containing bad pixel map.
 * @param rawData structure where raw data header information will be stored
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibLoadRawDataHdr(const char      *filename,
                                      amdlibRAW_DATA  *rawData,
                                      amdlibERROR_MSG errMsg)
{
    struct stat     statBuf;
    char            fitsioMsg[256];
    fitsfile        *filePtr;
    int             status = 0;
    char            keywName[64];
    int             i;
    double          dit;
    int             keysExist = 0;
    int             moreKeys = 0;
    amdlibKEYW_LINE record;

    amdlibLogTrace("amdlibLoadRawDataHdr()");
    
    /* If raw data structure is not initailized, do it */
    if (rawData->thisPtr != rawData)
    {
        amdlibInitRawData(rawData);
    }

    /* First free previous allocated memory */
    amdlibFreeRawData(rawData);

    /* Check the file exists */
    if (stat(filename, &statBuf) != 0)
    {
        amdlibSetErrMsg("File '%.80s' does not exist", filename);
        return amdlibFAILURE;
    }

    /* Open FITS file */
    if (fits_open_file(&filePtr, filename, READONLY, &status) != 0)
    {
        amdlibReturnFitsError(filename);
    }

    /* True if FITS file only contains detector data; i.e. does not contains
     * ICS sub-system nor wave data binary table*/
    rawData->detDataOnly = amdlibTRUE;

    /* Main header : get all keywords */

    /* Get the number of keywords in HDU */
    keysExist=0;
    moreKeys=0;
    if (fits_get_hdrspace(filePtr, &keysExist, &moreKeys,
                          &status) != 0)
    {
        status = 0;
    }

    /* For each keyword */
    amdlibClearInsCfg(&rawData->insCfg);
    for (i = 1; i <= keysExist; i++)
    {
        /* Read keyword line i */
        if (fits_read_record(filePtr, i, record, &status) != 0)
        {
            status = 0;
        }
        else
        {
            /* If it is not a simple comment */
            if (strstr(record, "COMMENT ") == NULL)
            {
                /* Store keyword */
                if (amdlibAddInsCfgKeyword(&rawData->insCfg,
                                           record, errMsg) == amdlibFAILURE)
                {
                    return amdlibFAILURE;
                }
            }

            /* If there is at least one instrument keyword, it is assumed that
             * ICS keywords have been merged to FITS header */
            if (strstr(record, "HIERARCH ESO INS") != NULL)
            {
                rawData->detDataOnly = amdlibFALSE;
            }
        }
    }

    /* Check the mandatory keywords are defined in header */
    if (rawData->detDataOnly == amdlibFALSE)
    {
        i = 0;
        while (amdlibRequiredInsCfgKeywList[i] != 0)
        {
            amdlibKEYWORD keyword;
            strcpy(keyword.name, amdlibRequiredInsCfgKeywList[i]);
            if (fits_read_key (filePtr, TSTRING, (char*)keyword.name,
                               keyword.value, NULL, &status) != 0)
            {
                amdlibGetFitsError(amdlibRequiredInsCfgKeywList[i])
                fits_close_file(filePtr, &status);
                return amdlibFAILURE;
            }
            i++;
        }
    }

    /* Main header : get P2VM id */
    strcpy (keywName, "HIERARCH ESO OCS P2VM ID");
    if (fits_read_key (filePtr, TINT, keywName, &rawData->p2vmId,
                       NULL, &status) != 0) 
    {
        /* No reference to P2VM */
        rawData->p2vmId = 0;
        status = 0;
    }

    /* Main header : get DIT */
    sprintf (keywName, "HIERARCH ESO DET DIT");
    if (fits_read_key (filePtr, TDOUBLE, (char*)keywName,
                       &dit, NULL, &status) != 0)
    {
        dit = 0.0;
        status = 0;
    }
    rawData->expTime = dit;

    /* Determine the frame type */
    if (rawData->detDataOnly == amdlibFALSE)
    {
        /* Main header : get frame type */
        char            frameTypeStr[64];
        strcpy (keywName, "HIERARCH ESO OCS DET FRAM TYPE");
        memset (frameTypeStr, '\0', sizeof(frameTypeStr));
        if (fits_read_key (filePtr, TSTRING, keywName, frameTypeStr,
                           NULL, &status) != 0) 
        {
            status = 0;
        }

        /* If OCS.DET.FRAM.TYPE keyword is specified, use it */
        if (strlen(frameTypeStr) != 0)
        {
            rawData->frameType = amdlibFrameTypeStr2Id(frameTypeStr);
        }
        else
        {
            amdlibSetErrMsg("Osolete input data - "
                            "no frame type specified in '%.80s'", filename);
            /*return amdlibFAILURE;*/
            rawData->frameType = amdlibUNKNOWN_FRAME;
        }
    }
    else
    {
        rawData->frameType = amdlibUNKNOWN_FRAME;
    }

    /* Read 'AMBER wave data' */
    if (rawData->detDataOnly == amdlibFALSE)
    {
        if (amdlibReadWaveData(filePtr,
                               &rawData->waveData, errMsg) != amdlibSUCCESS)
        {
            fits_close_file(filePtr, &status);
            return amdlibFAILURE;
        }
    }
    else
    {
        memset(&rawData->waveData, '\0', sizeof(amdlibWAVEDATA));
    }

    /* Read 'imaging detector header' */
    if (amdlibReadImagingDetectorHdr(filePtr, &rawData->imagingDetector,
                                    errMsg) != amdlibSUCCESS)
    {
        fits_close_file(filePtr, &status);
        return amdlibFAILURE;
    }

    /* Set the number of spectral rows and column according to
     * the number of regions
     */
    rawData->nbRegions = rawData->imagingDetector.nbRegions;
    switch(rawData->nbRegions)
    {
        case 1:
            rawData->nbRows=1;
            rawData->nbCols=1;
            break;
        case 4:
            rawData->nbRows=1;
            rawData->nbCols=4;
            break;
        case 8:
            rawData->nbRows=2;
            rawData->nbCols=4;
            break;
        case 12:
            rawData->nbRows=3;
            rawData->nbCols=4;
            break;
        case 5:
            rawData->nbRows=1;
            rawData->nbCols=5;
            break;
        case 10:
            rawData->nbRows=2;
            rawData->nbCols=5;
            break;
        case 15:
            rawData->nbRows=3;
            rawData->nbCols=5;
            break;
        default:
            fits_close_file(filePtr, &status);
            amdlibSetErrMsg("Number of regions '%d' inconsistant",
                            rawData->nbRegions);
            return amdlibFAILURE;

    }

    /* Read 'region informations' in imaging detector table */
    if (amdlibReadRegionInfo(filePtr, &rawData->region, rawData->nbRegions,
                            rawData->imagingDetector.maxTel,
                            errMsg) != amdlibSUCCESS)
    {
        fits_close_file(filePtr, &status);
        return amdlibFAILURE;
    }

    /* Read 'imaging data header' */
    if (amdlibReadImagingDataHdr(filePtr, &rawData->imagingData,
                                errMsg) != amdlibSUCCESS)
    {
        fits_close_file(filePtr, &status);
        return amdlibFAILURE;
    }

    /* Set the number of frames */
    rawData->nbFrames = rawData->region[0].dimAxis[2];

    /* Data is not yet loaded, and there is not calibrated */
    rawData->dataLoaded = amdlibFALSE;
    rawData->dataCalibrated = amdlibFALSE;

    /* Close the FITS file */
    if (fits_close_file(filePtr, &status) != 0)
    {
        amdlibReturnFitsError(filename);
    }
    return amdlibSUCCESS;
}

/**
 * Load N frames from raw data file.
 *
 * This function loads, as amdlibLoadRawData() function does, the raw data file,
 * but only the requested frames. The amdlibFIRST_FRAME macro can be used to
 * specify the first frame of the files, as well as the amdlibALL_FRAMES macro
 * to specify all remaining frames from the first specified frame to the last
 * one of the file.
 *
 * @param filename name of the FITS file containing raw data.
 * @param rawData structure where loaded information will be stored
 * @param firstFrame index of the first frame to load 
 * @param nbFrames number of frames to load 
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibLoadRawFrames(const char      *filename,
                                     amdlibRAW_DATA  *rawData,
                                     int             firstFrame,
                                     int             nbFrames,
                                     amdlibERROR_MSG errMsg)
{
    char            fitsioMsg[256];
    fitsfile        *filePtr;
    int             status = 0;

    amdlibLogTrace("amdlibLoadRawFrames()");
    
    /* First load header of raw file */
    if (amdlibLoadRawDataHdr(filename, rawData, errMsg) != amdlibSUCCESS)
    {
        return amdlibFAILURE;
    }

    /* Re-open FITS file */
    if (fits_open_file(&filePtr, filename, READONLY, &status) != 0)
    {
        amdlibReturnFitsError(filename);
    }

    /* Check if all frames have to be read */
    if (nbFrames == amdlibALL_FRAMES)
    {
        nbFrames = rawData->nbFrames - firstFrame + 1;
    }

    /* Read 'region data' in imaging data table */
    if (amdlibReadRegionData(filePtr, rawData->region,
                            rawData->nbRegions,
                            firstFrame, nbFrames,
                            errMsg) != amdlibSUCCESS)
    {
        fits_close_file(filePtr, &status);
        return amdlibFAILURE;
    }
    /* Set the number of frames */
    rawData->nbFrames = nbFrames;

    /* Read 'array geometry'  */
    if (amdlibReadArrayGeometry(filePtr, &rawData->arrayGeometry,
                                errMsg) != amdlibSUCCESS)
    {
        fits_close_file(filePtr, &status);
        return amdlibFAILURE;
    }
    
    /* Read 'time data' in imaging data table */
    if (amdlibReadTimeTag(filePtr, rawData,
                          firstFrame, nbFrames,
                          errMsg) != amdlibSUCCESS)
    {
        fits_close_file(filePtr, &status);
        return amdlibFAILURE;
    }
    
    /* Allocate memory to store bias */
    rawData->rmsBias = calloc(rawData->nbFrames, sizeof(float));
    if (rawData->rmsBias == NULL)
    {
        amdlibSetErrMsg("Could not allocate memory for RMS bias");
        return amdlibFAILURE;
    }
    /* Data is now loaded, but not yet calibrated */
    rawData->dataLoaded = amdlibTRUE;
    rawData->dataCalibrated = amdlibFALSE;

    /* Close the FITS file */
    if (fits_close_file(filePtr, &status) != 0)
    {
        amdlibReturnFitsError(filename);
    }
    return amdlibSUCCESS;
}

/**
 * Release memory allocated to store raw data and reset the structure members to
 * zero.
 *
 * @param rawData structure where raw data is stored
 */
void amdlibReleaseRawData(amdlibRAW_DATA *rawData)
{
    amdlibLogTrace("amdlibReleaseRawData()");
    
    amdlibFreeRawData(rawData);
    memset (rawData, '\0', sizeof(amdlibRAW_DATA));
}

/**
 * Split glued image into data regions
 *
 * This function split the glued image into regions. It is assumed that regions
 * is organised in rows and column, as shown below:
 * \code
 *  +----+-------+-------+------+
 * |    |       |       |      |
 * |    |       |       |      |
 * |    |       |       |      |
 * +----+-------+-------+------+
 * |    |       |       |      |
 * |    |       |       |      |
 * +----+-------+-------+------+
 * \endcode
 *
 * The detector configuration is given by the nbRows and nbCols parameters,
 * which respectively defines the number of rows and columns, and therefore the
 * number of regions. The dimensions of each regions are given by the colWidth
 * and rowHeight parameters.

 * @param gluedImage pointer to the glued image
 * @param nbRows number of rows
 * @param nbCols number of column
 * @param colWidth width of the regions for each column 
 * @param rowHeight height of the regions for each row 
 * @param rawData structure where raw data will be stored
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibGluedImage2RawData(float *gluedImage,
                                          int nbRows, int nbCols, 
                                          int colWidth[amdlibMAX_NB_COLS],
                                          int rowHeight[amdlibMAX_NB_ROWS],
                                          amdlibRAW_DATA *rawData,
                                          amdlibERROR_MSG errMsg)
{
    amdlibBOOLEAN detCfgChanged;
    int iRow, iCol, iY;
    int iRegion;
    int nbOfElements;
    int startPos;

    amdlibLogTrace("amdlibGluedImage2RawData()");
    
    /* If raw data structure is not initailized, do it */
    if (rawData->thisPtr != rawData)
    {
        amdlibInitRawData(rawData);
    }

    /* Check the detector configuration against region configuration */
    detCfgChanged = amdlibFALSE;
    if ((nbRows * nbCols) != rawData->nbRegions)
    {
        detCfgChanged = amdlibTRUE;
    }
    else
    {
        for (iRow=0; iRow < nbRows ; iRow++)
        {
            for (iCol = 0; iCol < nbCols; iCol++)
            {
                iRegion = iRow *  nbCols + iCol;
                if ((rawData->region[iRegion].dimAxis[0] != colWidth[iCol]) ||
                    (rawData->region[iRegion].dimAxis[1] != rowHeight[iRow]))
                {
                    detCfgChanged = amdlibTRUE;
                }
            }
        }
    }

    /* If detector configuration has changed, free and allocate new data
     * structure */
    if (detCfgChanged)
    {
        /* Free memory */
        amdlibFreeRawData(rawData);

        /* Allocate memory for region array */
        rawData->nbCols=nbCols;
        rawData->nbRows=nbRows;
        rawData->nbRegions=nbRows * nbCols;
        rawData->nbFrames = 1;
       /* Allocate memory for time Tag (1 value) */
        rawData->timeTag = calloc(rawData->nbFrames,sizeof(*(rawData->timeTag)));
        if (amdlibAllocateRegions
            (&rawData->region, rawData->nbRegions) != amdlibSUCCESS)
        {
            amdlibSetErrMsg("Could not allocate memory for regions");
            return amdlibFAILURE;
        }

        /* Allocate memory for regions */
        for (iRow = 0; iRow < nbRows ; iRow++)
        {
            for (iCol = 0; iCol < nbCols; iCol++)
            {
                iRegion = iRow * nbCols + iCol;
                nbOfElements = colWidth[iCol] * rowHeight[iRow];
                rawData->region[iRegion].data =
                calloc( nbOfElements, sizeof(float));
                if (rawData->region[iRegion].data == NULL)
                {
                    amdlibSetErrMsg("Could not allocate memory for data of "
                                    "region %d (size %dx%d)", iRegion,
                                    colWidth[iCol], rowHeight[iRow]);
                    return amdlibFAILURE;
                }
                rawData->region[iRegion].dimAxis[0]=colWidth[iCol];
                rawData->region[iRegion].dimAxis[1]=rowHeight[iRow];
                rawData->region[iRegion].dimAxis[2]=rawData->nbFrames;
            }
        }
        /* Allocate memory to store bias */
        rawData->rmsBias = calloc(rawData->nbFrames,sizeof(float));
        if (rawData->rmsBias == NULL)
        {
            amdlibSetErrMsg("Could not allocate memory for RMS bias");
            return amdlibFAILURE;
        }
    }

    /* Split glued image into region data */
    startPos = 0;
    for (iRow = 0; iRow < nbRows ; iRow++)
    {
        for (iY = 0; iY < rowHeight[iRow]; iY++)
        {
            for (iCol = 0; iCol < nbCols; iCol++)
            {
                iRegion = iRow *  nbCols + iCol;

                memcpy(&rawData->region[iRegion].data[iY*colWidth[iCol]],
                       &gluedImage[startPos], colWidth[iCol] * sizeof(float));
                startPos += colWidth[iCol];
            }
        }
    }
    /*Fill in a Time Tag*/
    rawData->timeTag[0] = time(NULL) / 86400.0; /*One frame*/

    /* Inform the data is loaded in structure, but there is not calibrated */
    rawData->dataLoaded = amdlibTRUE;
    rawData->dataCalibrated = amdlibFALSE;

    /* No reference to P2VM */
    rawData->p2vmId = 0;

    return amdlibSUCCESS;
}

/**
 * Calibrate raw data; i.e. compensate for detector effect 
 *
 * This function applies cosmetic corrections on raw data produced by the AMBER
 * instrument. This consists for each pixel to :
 *      \li discard null frames
 *      \li subtract the pixel bias (compensation for detector effects),
 *      \li subtract the row bias,
 *      \li divide the result by flat field,
 *      \li multiply the result by the gain to get photoelectron counts,
 *      \li subtract remanence.
 *
 * @param pixelBias pixel bias tob be substracted to raw data
 * @param rawData structure where raw data is stored
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned. */
amdlibCOMPL_STAT amdlibCalibrateRawData(amdlibPIXEL_BIAS_DATA *pixelBias,
                                        amdlibRAW_DATA        *rawData,
                                        amdlibERROR_MSG       errMsg)
{
    int   iFrame;
    int   iRow, iCol, iRegion;
    float rowBias[amdlibDET_SIZE_Y]; 
    float gain;
    int   nbGoodPixelsOnDetector = 0;
    int   nbSaturatedPixels = 0;

    amdlibLogTrace("amdlibCalibrateRawData()");
    
    /* Check raw data structure contains data */
    if (rawData->dataLoaded == amdlibFALSE)
    {
        amdlibSetErrMsg("The raw data structure does not "
                        "contain data. Check call to amdlibLoadRawData()");
        return amdlibFAILURE;
    }

    /* Check raw data has been already calibrated */
    if (rawData->dataCalibrated == amdlibTRUE)
    {
        amdlibSetErrMsg("The data has been already calibrated");
        return amdlibFAILURE;
    }

    /* Check compatibity of raw data and pixel bias map */
    if (amdlibCheckDataCompatibility(pixelBias, rawData, 
                                     errMsg) != amdlibSUCCESS)
    {
        return amdlibFAILURE;
    }
    
    /* Check for NULL Frames */
    if (amdlibDiscardNullFrames(rawData, errMsg) != amdlibSUCCESS)
    {
        return amdlibFAILURE;
    }

    /* Fix eventual start pix 0 to 1 in region header */
    for (iRow = 0; iRow < rawData->nbRows; iRow++)
    {
        for (iCol = 0; iCol < rawData->nbCols; iCol++)
        {
            iRegion = iRow * rawData->nbCols + iCol;
            
            if (rawData->region[iRegion].corner[0] == 0) 
            {
                rawData->region[iRegion].corner[0]  = 1;
            }
            if (rawData->region[iRegion].corner[1] == 0) 
            {
                rawData->region[iRegion].corner[1]  = 1;
            }
        }
    }
   
    /* Subtract Pixel Bias Map directly on raw data */
    if (amdlibSubtractPixelBias(pixelBias, rawData, errMsg) != amdlibSUCCESS)
    {
        return amdlibFAILURE;
    }
    
    /* For each frame */
    gain = rawData->region[0].gain;
    for (iFrame = 0; iFrame < rawData->region[0].dimAxis[2]; iFrame++)
    {
        /* Calculate row bias */        
        memset(rowBias, '\0', sizeof(rowBias));
        if (amdlibCalculateRowBias(rawData, iFrame, gain, 
                                   &nbGoodPixelsOnDetector, rowBias, errMsg) != 
            amdlibSUCCESS)
        {
            return amdlibFAILURE;
        }
        
        /* Calculate the calibrated image */
        nbSaturatedPixels = 0;
        if (amdlibCalibrateImage(rawData, iFrame, gain, rowBias,
                                 &nbSaturatedPixels, errMsg) != amdlibSUCCESS)
        {
            return amdlibFAILURE;
        }
      
        /* Check if detector is saturated; i.e. more than 1% of the pixels
        * higher than amdlibDETECTOR_SATURATION */
        if (nbSaturatedPixels >= (0.01 * nbGoodPixelsOnDetector))
        {
            rawData->dataIsSaturated = amdlibTRUE;
        }
        else
        {
            rawData->dataIsSaturated = amdlibFALSE;
        }
    }
    
    /* Subtract remanence */
    if (amdlibREMANENCE != 0)
    {
        if (amdlibSubtractRemanence(rawData, errMsg) != amdlibSUCCESS)
        {
            return amdlibFAILURE;
        }
    }

    /* Set 'calibrated' flag to TRUE */
    rawData->dataCalibrated = amdlibTRUE;

    return amdlibSUCCESS;
}

/** Max number of peaks along Y axis */
#define amdlibMAX_NB_SPEC 15
/** Max number of peaks along X axis */
#define amdlibMAX_NB_PEAKS 10

/**
 * Determine the positions of the columns
 *     
 * This function finds the position of the columns. This consists to find
 * the position, along the X axis, of the peak for each column. If several peaks
 * are found, this function computes the mean weighted position of the peaks. If
 * raw data has more than one row, the mean position of the peaks in the rows
 * for each column is returned. If no peak is found the column position is set
 * to 0.
 * This function returns, in addition to the column positions, the max value
 * found for each column.
 *
 * @param rawData structure where raw data is stored
 * @param colPos position of the peak found for each column. 
 * @param maxVal max value found for each column.
 *
 * @return
 * Always return amdlibSUCCESS
 */
amdlibCOMPL_STAT amdlibExtractColPos(amdlibRAW_DATA *rawData,
                                     float colPos[amdlibMAX_NB_COLS],
                                     float maxVal[amdlibMAX_NB_COLS])
{
    int col;
    int row;
    int iRegion;

    int x, y, frame;
    int nbX,nbY,nbFrames;

    double meanX[512];
    double pixelOffset[512];
    double mean;
    double variance;
    double threshold;

    amdlibLogTrace("amdlibExtractColPos()");
    
    /* Check raw data structure contains data */
    if (rawData->dataLoaded == amdlibFALSE)
    {
        return amdlibFAILURE;
    }

    /* Position of column 0 (i.e. dark) is not computed. Always set to 0 */
    colPos[0] = 0.0;

    /* For each column (excepted column 0 which is dark) */
    for(col=1; col<rawData->nbCols; col++)
    {
        double position[amdlibMAX_NB_ROWS][amdlibMAX_NB_PEAKS];
        double weight[amdlibMAX_NB_ROWS][amdlibMAX_NB_PEAKS];
        double max[amdlibMAX_NB_ROWS][amdlibMAX_NB_PEAKS];
        double meanPos;
        int    nbPos;

        /* For each row */
        for(row=0; row<rawData->nbRows; row++)
        {
            /* Get the region index and size */
            iRegion  = row*rawData->nbCols + col;
            nbX      = rawData->region[iRegion].dimAxis[0];
            nbY      = rawData->region[iRegion].dimAxis[1];
            nbFrames = rawData->region[iRegion].dimAxis[2];

            /* Compute mean value along X axis */
            for (x=0; x<nbX; x++)
            {
                double sum;

                sum  = 0.;
                pixelOffset[x] = rawData->region[iRegion].corner[0];
                for (y=0; y<nbY; y++)
                {
                    for (frame=0; frame<nbFrames; frame++)
                    {
                        sum += rawData->region[iRegion].data[frame*nbX*nbY+y*nbX+x];
                    }
                }
                meanX[x] = sum / (double)(nbY*nbFrames);
            }

            /* Compute the threshold used to find peaks according to the mean
             * value and variance of all pixels in region */
            mean      =0.0;
            variance  =0.0;
            threshold =0.0;
            for(x=0; x<nbX; x++)
            {
                mean     += meanX[x];
                variance += meanX[x]*meanX[x];
            }
            mean     /= (double)(nbX);
            variance -= (double)(nbX) * mean * mean ;
            variance  = variance /(double)(nbX-1);
            variance  = sqrt(variance );

            /* If the image is flat */
            if (variance < 10.0)
            {
                /* Do not search for peak */
                position[row][0] = 0.0;
            }
            else
            {

                threshold = mean + variance  ;   /* 4 sigma */
                nbPos=amdlibFindPeakPosition (meanX, nbX, threshold,
                                              amdlibMAX_NB_PEAKS,
                                              position[row], weight[row],
                                              max[row]);
                if (nbPos != 0)
                {
                    double w = 0.;
                    double wSum = 0.;
                    double m = 0.;
                    int pos;

                    for (pos=0; pos<nbPos; pos++)
                    {
                        w += weight[row][pos];
                        wSum += weight[row][pos] * position[row][pos];
                        if (m < max[row][pos])
                        {
                            m = max[row][pos];
                        }
                    }
                    position[row][0] = wSum/w;
                    max[row][0] = m;
                }
                else
                {
                    position[row][0] = 0.0;
                    max[row][0] = 0.0;
                }
            }
        }

        /* Compute the mean position and max value of the column */
        meanPos = 0.0;
        maxVal[col] = 0.0;
        nbPos = 0;
        for (row=0; row<rawData->nbRows; row++)
        {
            if (position[row][0] != 0.0)
            {
                meanPos += position[row][0];
                nbPos++;
                maxVal[col] += max[row][0];
            }
        }
        if (nbPos != 0)
        {
            meanPos /= nbPos;
            colPos[col] = pixelOffset[lround(meanPos)]+meanPos;
        }
        else
        {
            colPos[col] = 0.0;
        }
    }
    return amdlibSUCCESS;
}

/**
 * Determine the positions of the columns
 * Determine the position of the peak along Y axis
 *     
 * This function finds, for each region, the central spectral position. This
 * consists to find the position, along the Y axis, of the peak for each region.
 * If no peak is found the spectral position is set to 0.
 *
 * @param rawData structure where raw data is stored
 * @param spectPos position of the peak found for each region. 
 *
 * @return
 * Always return amdlibSUCCESS
 */
amdlibCOMPL_STAT amdlibExtractSpectPos(amdlibRAW_DATA *rawData,
                                       float spectPos[amdlibMAX_NB_COLS][amdlibMAX_NB_ROWS])
{
    int col;
    int row;
    int iRegion;

    int x, y, frame;
    int nbX,nbY,nbFrames;

    double meanY[512];
    double startPixel[512];
    double mean;
    double variance;
    double threshold;

    amdlibLogTrace("amdlibExtractSpectPos()");
    
    /* Check raw data structure contains data */
    if (rawData->dataLoaded == amdlibFALSE)
    {
        return amdlibFAILURE;
    }

    /* Position of column 0 (i.e. dark) is not computed. Always set to 0 */
    for(row=0; row<rawData->nbRows; row++)
    {
        spectPos[0][row] = 0;
    }

    /* For each column (excepted column 0 which is dark) */
    for(col=1; col<rawData->nbCols; col++)
    {
        int    nbPosition[amdlibMAX_NB_ROWS];
        double position[amdlibMAX_NB_ROWS][amdlibMAX_NB_SPEC];
        double weight[amdlibMAX_NB_ROWS][amdlibMAX_NB_SPEC];
        double max[amdlibMAX_NB_ROWS][amdlibMAX_NB_PEAKS]; /* Not used */

        /* For each row */
        memset(position, '\0', sizeof(position));
        for(row=0; row<rawData->nbRows; row++)
        {
            /* Get the region index and size */
            iRegion = row*rawData->nbCols + col;
            nbX      = rawData->region[iRegion].dimAxis[0];
            nbY      = rawData->region[iRegion].dimAxis[1];
            nbFrames = rawData->region[iRegion].dimAxis[2];

            /* Compute mean value along X axis */
            for (y=0; y<nbY; y++)
            {
                double sum;

                sum  = 0.;
                startPixel[y] = rawData->region[iRegion].corner[1];
                for (x=0; x<nbX; x++)
                {
                    for (frame=0; frame<nbFrames; frame++)
                    {
                        sum +=
                        rawData->region[iRegion].data[frame*nbX*nbY+y*nbX+x];
                    }
                }
                meanY[y] = sum / (double)(nbX*nbFrames);
            }
            /* Compute the threshold used to find peaks according to the mean
             * value and variance of all pixels in region */
            mean      =0.0;
            variance  =0.0;
            threshold =0.0;
            for(y=0; y<nbY; y++)
            {
                mean     += meanY[y];
                variance += meanY[y]*meanY[y];
            }
            mean      /= (double)(nbY);
            variance  -= (double)(nbY)*mean *mean ;
            variance  = variance /(double)(nbY-1);
            variance  = sqrt(variance );

            /* If the image is flat */
            if (variance < 10.0)
            {
                /* Do not search for peak */
                nbPosition[row] = 0;
            }
            else
            {
                threshold = mean + (2. * variance);   /* 4 sigma */
                nbPosition[row] = amdlibFindPeakPosition
                    (meanY, nbY, threshold, amdlibMAX_NB_SPEC, position[row],
                     weight[row], max[row]);
            }
        }

        /* Get the max peak positions on detector */
        for (row=0; row<rawData->nbRows; row++)
        {
            int i, iMax;
            double max;

            if (nbPosition[row] != 0)
            {
                /* Find max peak */
                iMax = 0;
                max = meanY[lround(position[row][iMax])];
                for (i=0;i<nbPosition[row];i++)
                {
                    if (meanY[lround(position[row][i])] >
                        meanY[lround(position[row][iMax])])
                    {
                        iMax = i;
                        max = meanY[lround(position[row][iMax])];
                    }
                }

                spectPos[col][row] = startPixel[lround(position[row][iMax])] +
                position[row][iMax];
            }
            else
            {
                spectPos[col][row] = 0.0;
            }
        }
    }
    return amdlibSUCCESS;
}

/** 
 * Check whether detector configurations for both raw data are the same or not. 
 *
 * This function returns true if the detector configurations (number of rows and
 * columns, and dimensions of each region) corresponding to the raw data are the
 * same. Otherwise, it returns false.
 *
 * @param rawData1 raw data containing first detector configuration
 * @param rawData2 raw data containing second detector configuration
 *
 * @return
 * true (amdlibTRUE) if configurations are the same, and false (amdlibFALSE)
 * otherwise.
 */
amdlibBOOLEAN amdlibIsSameDetCfg(amdlibRAW_DATA *rawData1,
                                 amdlibRAW_DATA *rawData2)
{
    amdlibBOOLEAN sameDetCfg;
    int iRow, iCol;
    int iRegion;

    amdlibLogTrace("amdlibIsSameDetCfg()");
    
    /* If raw data structure is not initailized, return FALSE */
    if (rawData1->thisPtr != rawData1)
    {
        return amdlibFALSE;
    }
    if (rawData2->thisPtr != rawData2)
    {
        return amdlibFALSE;
    }

    /* Compare the detector configurations */
    sameDetCfg = amdlibTRUE;
    if ((rawData1->nbRows != rawData2->nbRows) ||
        (rawData1->nbCols != rawData2->nbCols))
    {
        sameDetCfg = amdlibFALSE;
    }
    else
    {
        for (iRow=0; iRow < rawData1->nbRows ; iRow++)
        {
            for (iCol = 0; iCol < rawData1->nbCols; iCol++)
            {
                iRegion = iRow *  rawData1->nbCols + iCol;
                if ((rawData1->region[iRegion].corner[0] !=
                     rawData2->region[iRegion].corner[0]) ||
                    (rawData1->region[iRegion].corner[1] !=
                     rawData2->region[iRegion].corner[1]) ||
                    (rawData1->region[iRegion].dimAxis[0] !=
                     rawData2->region[iRegion].dimAxis[0]) ||
                    (rawData1->region[iRegion].dimAxis[1] !=
                     rawData2->region[iRegion].dimAxis[1]))
                {
                    sameDetCfg = amdlibFALSE;
                }
            }
        }
    }
    return sameDetCfg;
}

/*
 * Protected functions
 */
/**
 * Initialize raw data structure.
 *
 * @param rawData pointer to raw data structure
 */
void amdlibInitRawData(amdlibRAW_DATA *rawData)
{
    amdlibLogTrace("amdlibInitRawData()");
    
    /* Initialize data structure */
    memset (rawData, '\0', sizeof(amdlibRAW_DATA));
    rawData->thisPtr = rawData;
}

/**
 * Copy byte-by-byte the source file into the destination file.
 *
 * @param srcFilename name of the source file.
 * @param dstFilename name of the destination file.
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibCopyRawDataFile(const char      *srcFilename,
                                       const char      *dstFilename,
                                       amdlibERROR_MSG errMsg)
{
    FILE *source = NULL, *destination = NULL;
    int theChar;

    amdlibLogTrace("amdlibCopyRawDataFile()");
    
    /* Open files */
    source = fopen(srcFilename, "rb");
    if (source == NULL)
    {
        amdlibSetErrMsg("%s %s", srcFilename, strerror(errno));
        return amdlibFAILURE;
    }

    destination = fopen(dstFilename, "wb");
    if (destination == NULL)
    {
        amdlibSetErrMsg("%s %s", dstFilename, strerror(errno));
        fclose(source);
        return amdlibFAILURE;
    }

    /* Copie data from source file to destination file */
    while((theChar = fgetc(source)) != EOF)
    {
        if (fputc(theChar, destination) == EOF) break;
    }

    if (errno != 0)
    {
        amdlibSetErrMsg("%s", strerror(errno));
        return amdlibFAILURE;
    }

    /* Close file */
    fclose(source);
    fclose(destination);

    return amdlibSUCCESS;
}

/**
 * Store/Replace the detector data in the given file. 
 *
 * This function stores the detector data contained in the raw data structure
 * into the specified file. The specified file must be an existing raw data
 * file. This function just replaces the detector data contained in the
 * IMAGING_DATA binary table. 
 * 
 * @warning 
 * It does not change or update the information of the different headers, and
 * does not check correctness of these informations compared to the raw data.
 * Therefore, this function has to be only used to store, in the original file
 * or in a copy of the original file, the raw data which has been processed
 * using for example amdlibCalibrateRawData() function.
 * 
 * @param filename name of the FITS file containing raw data.
 * @param rawData structure where raw data is stored
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibStoreRawData(const char      *filename,
                                    amdlibRAW_DATA  *rawData,
                                    amdlibERROR_MSG errMsg)
{
    struct stat  statBuf;
    char         fitsioMsg[256];
    fitsfile     *filePtr;
    int          status = 0;

    amdlibLogTrace("amdlibStoreRawData()");
    
    /* Check the file exists */
    if (stat(filename, &statBuf) != 0)
    {
        amdlibSetErrMsg("File '%.80s' does not exist", filename);
        return amdlibFAILURE;
    }

    /* If raw data structure is not initailized, do it */
    if (rawData->thisPtr != rawData)
    {
        amdlibSetErrMsg("Raw data is not initialized");
        return amdlibFAILURE;
    }

    /* Open FITS file */
    if (fits_open_file(&filePtr, filename, READWRITE, &status) != 0)
    {
        amdlibReturnFitsError(filename);
    }

    /* Write 'region data' in imaging data table */
    if (amdlibWriteRegionData(filePtr, rawData->region,
                            rawData->nbRegions,
                            errMsg) != amdlibSUCCESS)
    {
        return amdlibFAILURE;
    }

    /* Write 'time tag' in imaging data table */
    if (amdlibWriteTimeTag(filePtr, rawData, errMsg) != amdlibSUCCESS)
    {
        return amdlibFAILURE;
    }

    /* Write the FITS file */
    if (fits_close_file(filePtr, &status) != 0)
    {
        amdlibReturnFitsError(filename);
    }

    return amdlibSUCCESS;
}

/**
 * Save raw data into a FITS file as 2-dimensional images. 
 *
 * This function creates glued images from data of the detector subwindows and
 * stores the resulting images into a normal FITS file.
 *
 * @param filename name of the FITS file where 2D images will be stored.
 * @param rawData structure where raw data is stored
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibSaveRawDataToFits(const char      *filename,
                                         amdlibRAW_DATA  *rawData,
                                         amdlibERROR_MSG errMsg)
{
    char         fitsioMsg[256];
    fitsfile     *filePtr;
    int          status = 0;
    int          i, iRegion;
    int          iRow, iCol;
    int          regionHeight, regionWidth, regionSize;
    int          iY, iX, iFrame;
    float        *image=NULL;

    int          bitpix = FLOAT_IMG;
    long         naxis = 3;  /* 2-dimensional image */
    long         naxes[3] = {0, 0, 0};

    amdlibLogTrace("amdlibSaveRawDataToFits()");
    
    /* If raw data structure is not initialized, do it */
    if (rawData->thisPtr != rawData)
    {
        amdlibSetErrMsg("Raw data not initialized");
        return amdlibFAILURE;
    }

    /* Check raw data structure contains data */
    if (rawData->dataLoaded == amdlibFALSE)
    {
        amdlibSetErrMsg("The raw data structure does not "
                        "contain data. Check call to amdlibLoadRawData()");
        return amdlibFAILURE;
    }

    /* Delete old file if it already exists */
    remove(filename);

    /* Create FITS file */
    if (fits_create_file(&filePtr, filename, &status) != 0)
    {
        amdlibReturnFitsError(filename);
    }

    /* Determine the size of the glued image */
    for (iCol = 0; iCol < rawData->nbCols; iCol++)
    {
        iRegion = iCol;
        naxes[0] += rawData->region[iRegion].dimAxis[0];
    }
    for (iRow = 0; iRow < rawData->nbRows; iRow++)
    {
        iRegion = iRow *  rawData->nbCols;
        naxes[1] += rawData->region[iRegion].dimAxis[1];
    }
    naxes[2] = rawData->region[0].dimAxis[2];

    /* Allocate memory */
    image = calloc(naxes[0]*naxes[1]*naxes[2],sizeof(float));
    if (image == NULL)
    {
        amdlibSetErrMsg("Could not allocate memory for glued image");
        return amdlibFAILURE;
    }

    /* Create glued image */
    i=0;
    for (iFrame = 0; iFrame < rawData->region[0].dimAxis[2]; iFrame++)
    {
        for (iRow = 0; iRow < rawData->nbRows; iRow++)
        {
            iRegion = iRow * rawData->nbCols;
            regionHeight = rawData->region[iRegion].dimAxis[1];
            for (iY = 0; iY < regionHeight; iY++)
            {
                for (iCol = 0; iCol < rawData->nbCols; iCol++)
                {
                    iRegion = iRow *  rawData->nbCols + iCol;
                    regionWidth = rawData->region[iRegion].dimAxis[0];
                    regionSize = regionWidth * regionHeight;
                    for (iX = 0; iX < regionWidth; iX++)
                    {
                        image[i++] =
                        rawData->region[iRegion].data[iFrame*regionSize +
                                                      iY*regionWidth + iX];
                    }
                }
            }
        }
    }

    /* Write the required keywords for the primary array image.     */
    if (fits_create_img(filePtr,  bitpix, naxis, naxes, &status) != 0)
    {
        free(image);
        amdlibGetFitsError(filename);
        fits_close_file(filePtr, &status);
        return amdlibFAILURE;
    }

    /* Copy array containing instrument configuration into file */
    for (i=0; i < rawData->insCfg.nbKeywords; i++)
    {
        amdlibKEYW_LINE keywLine;
        if (strstr(rawData->insCfg.keywords[i].name, "HIERARCH ESO") != NULL)
        {
            sprintf((char*)keywLine, "%s=%s/%s", 
                    rawData->insCfg.keywords[i].name,
                    rawData->insCfg.keywords[i].value,
                    rawData->insCfg.keywords[i].comment);
            if (fits_write_record(filePtr, keywLine, &status) != 0)
            {
                amdlibGetFitsError(rawData->insCfg.keywords[i].name);
                fits_close_file(filePtr, &status);
                return amdlibFAILURE;
            }
        }
    }

    /* write the array of unsigned integers to the FITS file */
    if (fits_write_img(filePtr, TFLOAT, 1, naxes[0]*naxes[1]*naxes[2], image,
                       &status) != 0)
    {
        free(image);
        amdlibGetFitsError(filename);
        fits_close_file(filePtr, &status);
        return amdlibFAILURE;
    }

    /* Release previously allocated memory */
    free(image);

    /* close file */
    if (fits_close_file(filePtr, &status) != 0)
    {
        amdlibReturnFitsError(filename);
    }
    return amdlibSUCCESS;
}

/**
 * Duplicate raw data structure. 
 *
 * This function allocates memory for destination data structure and copies the
 * data from source into it. 
 *
 * @param srcRawData source raw data structure.
 * @param dstRawData destination  raw data structure.
 * @param errMsg error description message returned if function fails.
  *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibDuplicateRawData(amdlibRAW_DATA  *srcRawData,
                                        amdlibRAW_DATA  *dstRawData,
                                        amdlibERROR_MSG errMsg)
{
    int i;
    
    amdlibLogTrace("amdlibDuplicateRawData()");
    
    /* Initialize destination data structure */
    if (dstRawData->thisPtr != dstRawData)
    {
        amdlibInitRawData(dstRawData);
    }
    
    /* Copy all raw data information */
    memcpy(dstRawData, srcRawData, sizeof(amdlibRAW_DATA));
      
    /* Allocates memory for regions */
    dstRawData->region = NULL;
    if (amdlibAllocateRegions(&dstRawData->region, 
                              dstRawData->nbRegions) != amdlibSUCCESS)
    {
        amdlibSetErrMsg("Could not allocate memory for regions");
        return amdlibFAILURE;
    }

    /* Copy all regions information and data */
    for (i = 0; i < dstRawData->nbRegions; i++)
    {
        int regionSize;
        /* Copy region information */
        memcpy(&dstRawData->region[i], &srcRawData->region[i], 
               sizeof(amdlibREGION));

        /* Allocate memory for this region */
        regionSize = srcRawData->region[i].dimAxis[0] * 
            srcRawData->region[i].dimAxis[1] * 
            srcRawData->region[i].dimAxis[2];
        dstRawData->region[i].data = calloc(regionSize, sizeof(float));
        if (dstRawData->region[i].data == NULL)
        {
            amdlibSetErrMsg("Could not allocate memory for data of region #%d",
                            i);
            return amdlibFAILURE;
        }
        
        /* Copy region data */
        memcpy(dstRawData->region[i].data, srcRawData->region[i].data, 
               regionSize * sizeof(float));
    }
    
    /* Copy time tag and computed rms bias for all frames */
    dstRawData->timeTag = calloc(dstRawData->nbFrames, sizeof(double));
    if (dstRawData->timeTag == NULL)
    {
        amdlibSetErrMsg("Could not allocate memory for time tag");
        return amdlibFAILURE;
    }
    dstRawData->rmsBias = calloc(dstRawData->nbFrames, sizeof(float));
    if (dstRawData->rmsBias == NULL)
    {
        amdlibSetErrMsg("Could not allocate memory for RMS bias");
        return amdlibFAILURE;
    }
    for (i = 0; i < dstRawData->nbFrames; i++)
    {
        dstRawData->timeTag[i] = srcRawData->timeTag[i];
        dstRawData->rmsBias[i] = srcRawData->rmsBias[i];
    }

    return amdlibSUCCESS;
}

/**
 * Sum and pack data of a channel 
 *
 * This function sums and packs (i.e. rows are merged together) data for the
 * given channel along X, Y and Z axis, according to the specified sum flag. The
 * result is stored in the array given as parameter. This array contains
 * pointers where each pointer points to the memory where result for frame i
 * has to be stored; i.e. result[i] contains pointer to memory allocated for
 * frame i.
 * 
 * @warning
 * The calling function has to allocate memory for the result.
 *
 * @param rawData raw data structure.
 * @param sumX flag indicating whether data has to be summed along X-axis or not
 * @param sumY flag indicating whether data has to be summed along Y-axis or not
 * @param sumZ flag indicating whether data has to be summed along Z-axis or not
 * @param channel channel number; could be amdlibPHOTO1_CHANNEL,
 * amdlibPHOTO2_CHANNEL, amdlibPHOTO3_CHANNEL or amdlibINTERF_CHANNEL
 * @param result array where result will be stored 
 * @param errMsg error description message returned if function fails.
  *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibSumAndPackData(amdlibRAW_DATA  *rawData,
                                      amdlibBOOLEAN    sumX,
                                      amdlibBOOLEAN    sumY,
                                      amdlibBOOLEAN    sumZ,
                                      int              channel,
                                      float            **result,
                                      amdlibERROR_MSG errMsg)
{
    int channelIndex;
    int dimX; /* resulting column width, ie number of pixels */
    int dimY; /* resulting column height, ie number of spectral channels */
    int dimZ; /* resulting column depth, ie number of frames */
    int iRow; /* index on rawData rows */
    int iRegion; /* index on regions */
    int currentY;
    int firstYInRow;
    int sX, sY, sZ;
    int iX, iY, iZ;
    float ***rawPtr; /* wrapping on rawData regions data */
    float **resPtr; /* wrapping on each result data frame */
    float **badPixels;

    amdlibLogTrace("amdlibSumAndPackData()");
    
    /* Find channel index */
    switch (channel)
    {
        case amdlibPHOTO1_CHANNEL :
        {
            channelIndex = 1;
            break;
        }
        case amdlibPHOTO2_CHANNEL :
        {
            channelIndex = 2;
            break;
        }
        case amdlibPHOTO3_CHANNEL :
        {
            channelIndex = 4;
            break;
        }
        case amdlibINTERF_CHANNEL :
        {
            channelIndex = 3;
            break;
        }
        default :
        {
            amdlibSetErrMsg("Channel reference '%d' does not exist", channel);
            return amdlibFAILURE;
        }
    }
    
    /* Find dimensions according to sumX, sumY and sumZ booleans */
    if (sumX == amdlibTRUE)
    {
        dimX = 1;
    }
    else
    {
        dimX = rawData->region[channelIndex].dimAxis[0]; 
    }
    if (sumY == amdlibTRUE)
    {
        dimY = 1;
    }
    else
    {
        dimY = 0;
        for (iRow = 0; iRow < rawData->nbRows; iRow ++)
        {
            dimY += 
                rawData->region[iRow*rawData->nbCols + channelIndex].dimAxis[1];
        }
    }
    if (sumZ == amdlibTRUE)
    {
        dimZ = 1;
    }
    else
    {
        dimZ = rawData->region[channelIndex].dimAxis[2];
    }
    /*  Reset result array */
    for (iZ = 0; iZ < dimZ; iZ ++)
    {
        if (result[iZ] == NULL)
        {
            amdlibSetErrMsg("Null pointer");
            return amdlibFAILURE;
        }
        else
        {
            memset(result[iZ], '\0', dimX * dimY * sizeof(float));
        }
    }
    
    /* Loop on rows so as to pack data */
    currentY = 0;
    for (iRow = 0; iRow < rawData->nbRows; iRow ++)
    {
        iRegion = iRow*rawData->nbCols + channelIndex;
        badPixels = amdlibGetBadPixelMapRegion(
                                    rawData->region[iRegion].corner[0] - 1,
                                    rawData->region[iRegion].corner[1] - 1,
                                    rawData->region[iRegion].dimAxis[0],
                                    rawData->region[iRegion].dimAxis[1],
                                    errMsg);
        if (badPixels == NULL)
        {
            return amdlibFAILURE;
        }
        rawPtr = amdlibWrap3DArrayFloat(rawData->region[iRegion].data, 
                                        rawData->region[iRegion].dimAxis[0],
                                        rawData->region[iRegion].dimAxis[1],
                                        rawData->region[iRegion].dimAxis[2],
                                        errMsg);
        if (rawPtr == NULL)
        {
            return amdlibFAILURE;
        }
        sX = 0;
        sY = 0;
        sZ = 0;
        firstYInRow = currentY;
        for (iZ = 0; iZ < rawData->region[iRegion].dimAxis[2]; iZ ++)
        {
            if (sumZ == amdlibFALSE)
            {
                sZ = iZ;
            }
            resPtr = amdlibWrap2DArrayFloat(result[sZ], dimX, dimY, errMsg);
            if (resPtr == NULL)
            {
                return amdlibFAILURE;
            }

            currentY = firstYInRow;
            for (iY = 0; iY < rawData->region[iRegion].dimAxis[1]; iY ++)
            {
                if (sumY == amdlibFALSE)
                {
                    sY = currentY;
                }
                for (iX = 0; iX < rawData->region[iRegion].dimAxis[0]; iX ++)
                {
                    if (sumX == amdlibFALSE)
                    {
                        sX = iX;
                    }
                    if (badPixels[iY][iX] == amdlibGOOD_PIXEL_FLAG)
                    {
                        resPtr[sY][sX] += rawPtr[iZ][iY][iX];
                    }
                }
                currentY ++;
            }
            amdlibFree2DArrayFloatWrapping(resPtr);
        }
        amdlibFree2DArrayFloat(badPixels);
        amdlibFree3DArrayFloatWrapping(rawPtr);        
    }
    
    /* If data has been summed along Z-axis (i.e. average of all frames)
     * computed the mean value. */
    if (sumZ == amdlibTRUE)
    {
        resPtr = amdlibWrap2DArrayFloat(result[0], dimX, dimY, errMsg);
        if (resPtr == NULL)
        {
            return amdlibFAILURE;
        }
        for (iY = 0; iY < dimY; iY++)
        {
            for (iX = 0; iX < dimX; iX ++)
            {
                resPtr[iY][iX] /= rawData->region[channelIndex].dimAxis[2];
            }
        }
        amdlibFree2DArrayFloatWrapping(resPtr);
    }

    return amdlibSUCCESS;
}

/**
 * Compute, sum and pack noise of a photometric channel 
 *
 * This function computes, sums and packs (i.e. rows are merged together) noise
 * for the given channel along X, Y and Z axis, according to the specified sum
 * flag. The result is stored in the array given as parameter. This array
 * contains pointers where each pointer points to the memory where result for
 * frame i has to be stored; i.e.  result[i] contains pointer to memory
 * allocated for frame i.
 * 
 * @warning
 * The calling function has to allocate memory for the result.
 *
 * @param rawData raw data structure.
 * @param sumX flag indicating whether data has to be summed along X-axis or not
 * @param sumY flag indicating whether data has to be summed along Y-axis or not
 * @param sumZ flag indicating whether data has to be summed along Z-axis or not
 * @param channel channel number; could amdlibPHOTO1_CHANNEL,
 * amdlibPHOTO2_CHANNEL or amdlibPHOTO3_CHANNEL
 * @param result array where result will be stored 
 * @param errMsg error description message returned if function fails.
  *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibSumAndPackSigma2PhotoData(amdlibRAW_DATA  *rawData,
                                                 amdlibBOOLEAN    sumX,
                                                 amdlibBOOLEAN    sumY,
                                                 amdlibBOOLEAN    sumZ,
                                                 int              channel,
                                                 float            **result,
                                                 amdlibERROR_MSG errMsg)
{
    int channelIndex;
    int dimX; /* resulting column width, ie number of pixels */
    int dimY; /* resulting column height, ie number of spectral channels */
    int dimZ; /* resulting column depth, ie number of frames */
    int iRow; /* index on rawData rows */
    int iRegion; /* index on regions */
    int currentY;
    int firstYInRow;
    int sX, sY, sZ;
    int iX, iY, iZ;
    float ***rawPtr; /* wrapping on rawData regions data */
    float **resPtr; /* wrapping on each frame of result data */
    float **badPixels;
    float **flatField;
    float sigma2Det;    

    amdlibLogTrace("amdlibSumAndPackSigma2PhotoData()");
    
    /* Find channel index */
    switch (channel)
    {
        case amdlibPHOTO1_CHANNEL :
        {
            channelIndex = 1;
            break;
        }
        case amdlibPHOTO2_CHANNEL :
        {
            channelIndex = 2;
            break;
        }
        case amdlibPHOTO3_CHANNEL :
        {
            channelIndex = 4;
            break;
        }
        default :
        {
            amdlibSetErrMsg("Channel reference '%d' does not exist", channel);
            return amdlibFAILURE;
        }
    }
    /* Find dimensions according to sumX, sumY and sumZ booleans */
    if (sumX == amdlibTRUE)
    {
        dimX = 1;
    }
    else
    {
        dimX = rawData->region[channelIndex].dimAxis[0]; 
    }
    if (sumY == amdlibTRUE)
    {
        dimY = 1;
    }
    else
    {
        dimY = 0;
        for (iRow = 0; iRow < rawData->nbRows; iRow ++)
        {
            dimY += 
                rawData->region[iRow*rawData->nbCols + channelIndex].dimAxis[1];
        }
    }
    if (sumZ == amdlibTRUE)
    {
        dimZ = 1;
    }
    else
    {
        dimZ = rawData->region[channelIndex].dimAxis[2];
    }
    /*  Reset result array */
    for (iZ = 0; iZ < dimZ; iZ ++)
    {
        if (result[iZ] == NULL)
        {
            amdlibSetErrMsg("Null pointer");
            return amdlibFAILURE;
        }
        else
        {
            memset(result[iZ], '\0', dimX * dimY * sizeof(float));
        }
    }

    /* Loop on rows so as to pack data */
    currentY = 0;
    for (iRow = 0; iRow < rawData->nbRows; iRow ++)
    {
        iRegion = iRow*rawData->nbCols + channelIndex;
        badPixels = amdlibGetBadPixelMapRegion(
                                    rawData->region[iRegion].corner[0] - 1,
                                    rawData->region[iRegion].corner[1] - 1,
                                    rawData->region[iRegion].dimAxis[0],
                                    rawData->region[iRegion].dimAxis[1],
                                    errMsg);
        if (badPixels == NULL)
        {
            return amdlibFAILURE;
        }
        flatField = amdlibGetFlatFieldMapRegion(
                                    rawData->region[iRegion].corner[0] - 1,
                                    rawData->region[iRegion].corner[1] - 1,
                                    rawData->region[iRegion].dimAxis[0],
                                    rawData->region[iRegion].dimAxis[1],
                                    errMsg);
        if (flatField == NULL)
        {
            return amdlibFAILURE;
        }
        rawPtr = amdlibWrap3DArrayFloat(rawData->region[iRegion].data, 
                                        rawData->region[iRegion].dimAxis[0],
                                        rawData->region[iRegion].dimAxis[1],
                                        rawData->region[iRegion].dimAxis[2],
                                        errMsg);
        if (rawPtr == NULL)
        {
            return amdlibFAILURE;
        }
        sX = 0;
        sY = 0;
        sZ = 0;
        firstYInRow = currentY;
        for (iZ = 0; iZ < rawData->region[iRegion].dimAxis[2]; iZ ++)
        {
            sigma2Det = amdlibPow2(rawData->rmsBias[iZ]);
            if (sumZ == amdlibFALSE)
            {
                sZ = iZ;
            }
            resPtr = amdlibWrap2DArrayFloat(result[sZ], dimX, dimY, errMsg);
            if (resPtr == NULL)
            {
                return amdlibFAILURE;
            }
            currentY = firstYInRow;
            for (iY = 0; iY < rawData->region[iRegion].dimAxis[1]; iY ++)
            {
                if (sumY == amdlibFALSE)
                {
                    sY = currentY;
                }
                for (iX = 0; iX < rawData->region[iRegion].dimAxis[0]; iX ++)
                {
                    if (sumX == amdlibFALSE)
                    {
                        sX = iX;
                    }
                    if (badPixels[iY][iX] == amdlibGOOD_PIXEL_FLAG)
                    {
                        resPtr[sY][sX] += 
                            (rawPtr[iZ][iY][iX] / flatField[iY][iX]) + 
                            sigma2Det / amdlibPow2(flatField[iY][iX]);
                    }
                }
                currentY ++;
            }
            amdlibFree2DArrayFloatWrapping(resPtr);
        }
        amdlibFree2DArrayFloat(badPixels);
        amdlibFree2DArrayFloat(flatField);
        amdlibFree3DArrayFloatWrapping(rawPtr);        
    }
    
    if (sumZ == amdlibTRUE)
    {
        resPtr = amdlibWrap2DArrayFloat(result[0], dimX, dimY, errMsg);
        if (resPtr == NULL)
        {
            return amdlibFAILURE;
        }
        for (iY = 0; iY < dimY; iY++)
        {
            for (iX = 0; iX < dimX; iX ++)
            {
                resPtr[iY][iX] /= rawData->region[channelIndex].dimAxis[2];
            }
        }
        amdlibFree2DArrayFloatWrapping(resPtr);
    }

    return amdlibSUCCESS;

}

/*
 * Local functions
 */
/**
 * Get time tags for the list of frames. 
 *
 * This function retrieves the time tags from the given FITS file for the
 * specified frames.
 * 
 * @param filePtr pointer to the FITS file containing raw data.
 * @param rawData structure where raw data is stored
 * @param firstFrame index of the first frame to load 
 * @param nbFrames number of frames to load 
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibReadTimeTag (fitsfile        *filePtr, 
                                    amdlibRAW_DATA  *rawData,
                                    int             firstFrame,
                                    int             nbFrames,
                                    amdlibERROR_MSG errMsg)
{
    int        status = 0;
    int        anynull = 0;
    char       fitsioMsg[256];
    int        colnum;
    
    amdlibLogTrace("amdlibReadTimeTag()");
    
    /* Reset error message */
    memset(errMsg, '\0', sizeof(amdlibERROR_MSG));

    /* Allocate memory to store time tag list */
    rawData->timeTag = calloc(nbFrames, sizeof(*(rawData->timeTag)));
    if (rawData->timeTag == NULL)
    {
        amdlibSetErrMsg("Could not allocate memory for time tag list");
        return amdlibFAILURE;
    }

   /* Go to the IMAGING_DATA binary table */
    if (fits_movnam_hdu(filePtr, BINARY_TBL, 
                        "IMAGING_DATA", 0, &status) != 0)
    {
        amdlibReturnFitsError("IMAGING_DATA");
    }

    if(fits_get_colnum (filePtr, CASEINSEN, "TIME", &colnum, &status))
    {
        amdlibReturnFitsError("TIME");
    }

    if(fits_read_col (filePtr, TDOUBLE, colnum, firstFrame, 1, nbFrames,
                      NULL, rawData->timeTag, &anynull, &status) != 0)
    {
        amdlibReturnFitsError("TIME");
    }
    return amdlibSUCCESS;
}

/**
 * Set time tags for all frames. 
 *
 * This function write the time tags stored in the raw data structure into the
 * FITS file.
 * 
 * @param filePtr pointer to the FITS file containing raw data.
 * @param rawData structure where raw data is stored
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibWriteTimeTag (fitsfile        *filePtr, 
                                     amdlibRAW_DATA  *rawData,
                                     amdlibERROR_MSG errMsg)
{
    int        status = 0;
    char       fitsioMsg[256];
    int        colnum;

    amdlibLogTrace("amdlibWriteTimeTag()");
    
    /* Reset error message */
    memset(errMsg, '\0', sizeof(amdlibERROR_MSG));

    /* Go to the IMAGING_DATA binary table */
    if (fits_movnam_hdu(filePtr, BINARY_TBL, "IMAGING_DATA", 0, &status) != 0)
    {
        amdlibReturnFitsError("IMAGING_DATA");
    }

    if(fits_get_colnum (filePtr, CASEINSEN, "TIME", &colnum, &status))
    {
        amdlibReturnFitsError("TIME");
    }
    /* Check data pointer */
    if (rawData->timeTag == NULL)
    {        
        amdlibSetErrMsg("The pointer to the time data is invalid"); 
        return amdlibFAILURE;
    }

    /* Write the time data */
    if (fits_write_col(filePtr, TDOUBLE, colnum, 1, 1, 
                       rawData->nbFrames, rawData->timeTag, &status) != 0)
    {
        amdlibReturnFitsError("TIME");
    }
    return amdlibSUCCESS;
}

/** 
 * Macro to sort two pixel values.
 */
#define amdlibPIX_SORT(a,b) { if ((a)>(b)) amdlibPIX_SWAP((a),(b)); }
/** 
 * Macro to swap two pixel values.
 */
#define amdlibPIX_SWAP(a,b) { float tmp=(a);(a)=(b);(b)=tmp; }
/**
 * Compute the median. 
 *
 * This function computes the median values of the entries in the given table.
 * It is superfast (since completely hardcoded median filter for 9 values).
 *
 * @note
 * Formula from: XILINX XCELL magazine, vol. 23 by John L. Smith
 * 
 * @param p table containing the 9 values.
 *
 * @return
 * The median value.
 */
float amdlibMedian9(float *p)
{
    amdlibPIX_SORT(p[1], p[2]); amdlibPIX_SORT(p[4], p[5]);
    amdlibPIX_SORT(p[7], p[8]); amdlibPIX_SORT(p[0], p[1]); 
    amdlibPIX_SORT(p[3], p[4]); amdlibPIX_SORT(p[6], p[7]); 
    amdlibPIX_SORT(p[1], p[2]); amdlibPIX_SORT(p[4], p[5]); 
    amdlibPIX_SORT(p[7], p[8]); amdlibPIX_SORT(p[0], p[3]); 
    amdlibPIX_SORT(p[5], p[8]); amdlibPIX_SORT(p[4], p[7]);
    amdlibPIX_SORT(p[3], p[6]); amdlibPIX_SORT(p[1], p[4]); 
    amdlibPIX_SORT(p[2], p[5]); amdlibPIX_SORT(p[4], p[7]); 
    amdlibPIX_SORT(p[4], p[2]); amdlibPIX_SORT(p[6], p[4]);
    amdlibPIX_SORT(p[4], p[2]); 
    return(p[4]) ;
}
#undef amdlibPIX_SORT
#undef amdlibPIX_SWAP

/**
 * Free memory allocated memory of raw data structure.
 *
 * This function frees previously allocated memory (if any) where raw data has
 * been stored. 
 * 
 * @param rawData structure where raw data is stored
 */
void amdlibFreeRawData(amdlibRAW_DATA *rawData)
{
    amdlibLogTrace("amdlibFreeRawData()");
    
    amdlibFreeRegions(&rawData->region, rawData->nbRegions);
    amdlibReleaseOiArray(&rawData->arrayGeometry);
    if (rawData->timeTag != NULL)
    {
        free(rawData->timeTag);
        rawData->timeTag = NULL;
    }
    if (rawData->rmsBias != NULL)
    {
        free(rawData->rmsBias);
        rawData->rmsBias = NULL;
    }
}

/**
 * Convert string frame type to Id.
 *
 * This function converts the string corresponding the the frame as defined by
 * OCS.DET.FRAM.TYPE keyword to the frame type Id as defined by the
 * amdlibFRAME_TYPE enumerate.
 *
 * @param frameTypeStr frame type as string 
 *
 * @return
 * Frame type Id.
 */
amdlibFRAME_TYPE amdlibFrameTypeStr2Id(char *frameTypeStr)
{
    amdlibLogTrace("amdlibFrameTypeStr2Id()");
    
    if (strcmp(frameTypeStr, "DARK") == 0)
    {
        return amdlibDARK_FRAME;
    }
    if ((strcmp(frameTypeStr, "TEL1") == 0) || 
        (strcmp(frameTypeStr, "TEL1SC") == 0))
    {
        return amdlibTEL1_FRAME;
    }
    if ((strcmp(frameTypeStr, "TEL2") == 0) ||
        (strcmp(frameTypeStr, "TEL2SC") == 0))
    {
        return amdlibTEL2_FRAME;
    }
    if ((strcmp(frameTypeStr, "TEL3") == 0) ||
        (strcmp(frameTypeStr, "TEL3SC") == 0))
    {
        return amdlibTEL3_FRAME;
    }
    if (strcmp(frameTypeStr, "TEL12") == 0)
    {
        return amdlibTEL12_FRAME;
    }
    if (strcmp(frameTypeStr, "TEL12Q") == 0)
    {
        return amdlibTEL12Q_FRAME;
    }
    if (strcmp(frameTypeStr, "TEL23") == 0)
    {
        return amdlibTEL23_FRAME;
    }
    if (strcmp(frameTypeStr, "TEL23Q") == 0)
    {
        return amdlibTEL23Q_FRAME;
    }
    if (strcmp(frameTypeStr, "TEL13") == 0)
    {
        return amdlibTEL13_FRAME;
    }
    if (strcmp(frameTypeStr, "TEL13Q") == 0)
    {
        return amdlibTEL13Q_FRAME;
    }

    return amdlibUNKNOWN_FRAME;
}

/**
 * Find peaks in the given value array. 
 *
 * This function finds peaks (i.e. greater than the given threshold value),
 * and compute the weight of each peak.
 *
 * @param data table of values.
 * @param nbData number of values in table. 
 * @param threshold value used to detect peaks 
 * @param nbMaxPeaks number of peaks which can be found; i.e. size of the two
 * following arrays.
 * @param peakPos position of the found peaks
 * @param peakWeight weight of the found peaks
 * @param peakMax maximum value of the found peaks
 *
 * @return
 * The number of the found peaks
 */
int amdlibFindPeakPosition(double *data, int nbData, double threshold,
                           int nbMaxPeaks, double *peakPos, double *peakWeight,
                           double *peakMax)
{
    int    i, nbPeaks = 0;
    double delta;
    double weight = 0.;
    double weightedSum = 0.;
    double max = 0.;

    amdlibBOOLEAN peakFound = amdlibFALSE;

    amdlibLogTrace("amdlibFindPeakPosition()");
    
#if 0
    /* Print out command lines to visualize data using yorick program */
    amdlibLogTrace("data=[");
    for (i = 0; i < nbData; i++)
    {
        amdlibLogTrace("%.1f",  data[i]);
        if(i!=(nbData-1))
            amdlibLogTrace(",");
    }
    amdlibLogTrace("]");
    amdlibLogTrace("thr=array(float(%.1f), dimsof(data)(2))", threshold);
    amdlibLogTrace("plg,data");
    amdlibLogTrace("plg,thr,color=\"red\"");
#endif

    /* For each value in table */
    for (i = 0; i < nbData; i++)
    {
        delta = data[i] - threshold;

        /* If value is greater than threshold */
        if (delta > 0.0)
        {
            /* Update weighted sum and weight */
            weight      += delta;
            weightedSum += i * delta;
            peakFound   = amdlibTRUE;
            /* Update max value */
            if (max < data[i])
            {
                max = data[i];
            }
        }
        /* End if */

        /* If value is lower than threshold (or it is the last value) and peak
         * has been found, corresponding to the end of peak */
        if (((delta <= 0.) || (i == (nbData-1))) && (peakFound == amdlibTRUE))
        {
            /* Compute position of peak (barycenter) */
            peakPos[nbPeaks]    = weightedSum / weight;
            peakWeight[nbPeaks] = weight;
            peakMax[nbPeaks]    = max;

            nbPeaks++;
            /* If max number of peaks is reached */
            if(nbPeaks >= nbMaxPeaks)
            {
                /* Stop */
                return nbPeaks;
            }
            /* End if */

            weightedSum = 0.0;
            weight = 0.0;
            max = 0.0;
            peakFound = amdlibFALSE;
        }
        /* End if */
    }
    /* End for */
    return nbPeaks;
}

/**
 * Check pixel bias map and data compatibity
 *
 * This function checks if data from the pixel bias map and data from the raw 
 * data structure are compatible, i.e. if each field stored in both structures
 * is instanciated twice by the same value.
 *
 * @param pixelBias pixel bias map.
 * @param rawData raw data. 
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibCheckDataCompatibility(amdlibPIXEL_BIAS_DATA *pixelBias,
                                              amdlibRAW_DATA        *rawData,
                                              amdlibERROR_MSG       errMsg)
{
    int iRegion;
    
    amdlibLogTrace("amdlibCheckDataCompatibility()");
    
    if (pixelBias != NULL)
    {
        if (pixelBias->nbRegions != rawData->nbRegions)
        {
            amdlibSetErrMsg("The number of regions in raw data (%d) and in "
                            "pixel bias map (%d) are incompatible",
                            rawData->nbRegions, pixelBias->nbRegions);
            return amdlibFAILURE;
        }

        if (pixelBias->expTime != rawData->expTime)
        {
            /* Use a Warning this time */
            amdlibLogWarning("The expTime in raw data (%f) and in pixel bias "
                             "map (%f) are incompatible",
                             rawData->expTime, pixelBias->expTime);
        }

        if (pixelBias->nbRows != rawData->nbRows)
        {
            amdlibSetErrMsg("The number of rows in raw data (%d) and in "
                            "pixel bias map (%d) are incompatible",
                            rawData->nbRows, pixelBias->nbRows);
            return amdlibFAILURE;
        }

        if (pixelBias->nbCols != rawData->nbCols)
        {
            amdlibSetErrMsg("The number of columns in raw data (%d) and in "
                            "pixel bias map (%d) are incompatible",
                            rawData->nbCols, pixelBias->nbCols);
            return amdlibFAILURE;
        }

        /* For each region */
        for (iRegion = 0; iRegion < rawData->nbRegions; iRegion++)
        {
            /* Check the region position */
            if (pixelBias->region[iRegion].corner[0] !=
                rawData->region[iRegion].corner[0])
            {
                amdlibSetErrMsg("The position in X direction of the region %d "
                                "in raw data (%d) and in pixel bias map (%d) "
                                "are incompatible",
                                iRegion, rawData->region[iRegion].corner[0],
                                pixelBias->region[iRegion].corner[0]);
                return amdlibFAILURE;
            }

            if (pixelBias->region[iRegion].corner[1] !=
                rawData->region[iRegion].corner[1])
            {
                amdlibSetErrMsg("The position in X direction of the region %d "
                                "in raw data (%d) and in pixel bias map (%d) "
                                "are incompatible",
                                iRegion, rawData->region[iRegion].corner[1],
                                pixelBias->region[iRegion].corner[1]);
                return amdlibFAILURE;
            }

            /* Check the region size */
            if (pixelBias->region[iRegion].dimAxis[0] !=
                rawData->region[iRegion].dimAxis[0])
            {
                amdlibSetErrMsg("The number of pixels  in X direction of the "
                                "region %d in raw data (%d) and in pixel bias "
                                "map (%d are incompatible",
                                iRegion, rawData->region[iRegion].dimAxis[0],
                                pixelBias->region[iRegion].dimAxis[0]);
                return amdlibFAILURE;
            }
            if (pixelBias->region[iRegion].dimAxis[1] !=
                rawData->region[iRegion].dimAxis[1])
            {
                amdlibSetErrMsg("The number of pixels in X direction of the "
                                "region %d in raw data (%d) and in pixel bias "
                                "map (%d) are incompatible",
                                iRegion, rawData->region[iRegion].dimAxis[1],
                                pixelBias->region[iRegion].dimAxis[1]);
                return amdlibFAILURE;
            }
        }
    }
    return amdlibSUCCESS;
}

/**
 * Delete null frames 
 *
 * This function treats the problem of null frames present in the raw data being
 * calibrated. These null frames are not taken into account, they are shifted at
 * the end of the list of frames and ignored.
 *
 * @param rawData raw data. 
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibDiscardNullFrames(amdlibRAW_DATA        *rawData,
                                         amdlibERROR_MSG       errMsg)
{
    int nbFrames;
    int hasNullFrames;
    int nbNullFrames;
    int iFrame;
    int onbFrames;
    int oFrame;
    int iRegion;
    int regionSize;
    
    amdlibLogTrace("amdlibDiscardNullFrames()");
    
    nbFrames = rawData->region[0].dimAxis[2];
    hasNullFrames = 0;
    nbNullFrames = 0;
    for (iFrame = 0; iFrame < nbFrames ; iFrame++)
    {
        if (rawData->timeTag[iFrame] == 0.0)
        {
            /* Use a Warning Only */
            if (hasNullFrames == 0) 
            {
                amdlibLogWarning("Data has null frames");
                hasNullFrames = 1;
            }
            nbNullFrames++;
        }
    }

    /* Compact non-null frames and update structure header accordingly*/
    if (hasNullFrames) 
    {
        /* Report on Null Frame Number */
        amdlibLogWarningDetail("Lost %.2f%% of %d frames due to DCS error",
                               nbNullFrames*100.0/(float)nbFrames, nbFrames);
        onbFrames = nbFrames;
        /* Copy back each non-null frame values and infos after the last good */
        for (iFrame = 0, oFrame = 0; iFrame < nbFrames; iFrame++)
        {
            if (rawData->timeTag[iFrame] == 0.0)
            {
                onbFrames--;
            }
            else
            {
                if (iFrame != oFrame)
                {
                    for (iRegion = 0; iRegion < rawData->nbRegions; iRegion++)
                    {
                        regionSize = rawData->region[iRegion].dimAxis[0] *
                                     rawData->region[iRegion].dimAxis[1];
                        memcpy(
                          rawData->region[iRegion].data + oFrame*regionSize,
                          rawData->region[iRegion].data + iFrame*regionSize,
                          regionSize*sizeof(*(rawData->region[iRegion].data)));
                    }
                    rawData->timeTag[oFrame] = rawData->timeTag[iFrame];
                }
                oFrame++;
            }
        }
        if (onbFrames == 0)
        {
            amdlibSetErrMsg("Only Null Frames found !");
            return amdlibFAILURE;
        }
        
        /* Update nb frame info where it hurts... */
        rawData->nbFrames = onbFrames;
        for (iRegion = 0; iRegion < rawData->nbRegions; iRegion++)
        {
            rawData->region[iRegion].dimAxis[2] = onbFrames;
        }
        nbFrames = onbFrames;
    }
    
    return amdlibSUCCESS;
}

/**
 * Subtracts pixel bias
 *
 * This function subtracts pixel bias values directly on raw data.
 *
 * @param pixelBias pixel bias map.
 * @param rawData raw data. 
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibSubtractPixelBias(amdlibPIXEL_BIAS_DATA *pixelBias,
                                         amdlibRAW_DATA        *rawData,
                                         amdlibERROR_MSG       errMsg)
{
    int iRegion;
    int iFrame;
    int iX, iY;
    float ***rawPtr;
    float **pbmPtr;    
    
    amdlibLogTrace("amdlibSubtractPixelBias()");
    
    if (pixelBias != NULL)
    {
        for (iRegion = 0; iRegion < rawData->nbRegions; iRegion++)
        {
            rawPtr = 
                amdlibWrap3DArrayFloat(rawData->region[iRegion].data, 
                                       rawData->region[iRegion].dimAxis[0],
                                       rawData->region[iRegion].dimAxis[1],
                                       rawData->region[iRegion].dimAxis[2],
                                       errMsg);
            pbmPtr = 
                amdlibWrap2DArrayFloat(pixelBias->region[iRegion].data, 
                                       pixelBias->region[iRegion].dimAxis[0],
                                       pixelBias->region[iRegion].dimAxis[1],
                                       errMsg);
            if (rawPtr == NULL) 
            {
                return amdlibFAILURE;
            }
            if (pbmPtr == NULL)
            {
                amdlibFree3DArrayFloatWrapping(rawPtr);
                return amdlibFAILURE;
            }
            for (iFrame = 0; iFrame < rawData->region[iRegion].dimAxis[2]; 
                 iFrame++)
            {
                for (iY = 0; iY < rawData->region[iRegion].dimAxis[1]; iY++)
                {
                    for (iX = 0; iX < rawData->region[iRegion].dimAxis[0]; iX++)
                    {
                        rawPtr[iFrame][iY][iX] -= pbmPtr[iY][iX];
                    }
                }
            }
            amdlibFree3DArrayFloatWrapping(rawPtr);
            amdlibFree2DArrayFloatWrapping(pbmPtr);
        }
    }
    return amdlibSUCCESS;
}

#define amdlibSTART_MASK_COL 5
/**
 * Calculate the row bias so as to normalise data
 *
 * This function calculates the row bias for the frame i. 
 *
 * @param rawData raw data. 
 * @param iFrame index of the frame 
 * @param gain gain to convert photoelectron counts
 * @param nbGoodPixelsOnDetector number of found good pixels
 * @param rowBias array where computed bias for each line is stored
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibCalculateRowBias(
                              amdlibRAW_DATA        *rawData,
                              int                   iFrame,
                              float                 gain,
                              int                   *nbGoodPixelsOnDetector,
                              float                 *rowBias,
                              amdlibERROR_MSG       errMsg)
{
    float rmsBias, finalRmsBias;
    int   nbGoodPixelsInRow[amdlibDET_SIZE_Y];
    int   iRow;
    int   iRegion;
    int   regionSize;
    int   line;
    int   iX, iY;
    float **rawPtr;
    float **badPixels;
    float sortArray[amdlibDET_SIZE_X];
    float median;
    float sum;

    amdlibLogTrace("amdlibCalculateRowBias()");
    
    memset(nbGoodPixelsInRow, '\0', sizeof(nbGoodPixelsInRow));
    rmsBias = 0.0;
    *nbGoodPixelsOnDetector = 0;
    for (iRow = 0; iRow < rawData->nbRows; iRow++)
    {
        /* Take first region of each row corresponding to the shielded area */
        iRegion = iRow * rawData->nbCols;
        regionSize = rawData->region[iRegion].dimAxis[0] *
            rawData->region[iRegion].dimAxis[1];
        rawPtr = amdlibWrap2DArrayFloat
            (&rawData->region[iRegion].data[regionSize*iFrame], 
             rawData->region[iRegion].dimAxis[0],
             rawData->region[iRegion].dimAxis[1],
             errMsg);
        if (rawPtr == NULL)
        {
            return amdlibFAILURE;
        }
        /* Take the part of the bad pixel map corresponding to the current 
         * region */
        badPixels = amdlibGetBadPixelMapRegion(
                        rawData->region[iRegion].corner[0] - 1,
                        rawData->region[iRegion].corner[1] - 1,
                        rawData->region[iRegion].dimAxis[0],
                        rawData->region[iRegion].dimAxis[1],
                        errMsg);
        if (badPixels == NULL)
        {
            amdlibFree2DArrayFloatWrapping(rawPtr);
            return amdlibFAILURE;
        }
        for (iY = 0; iY <  rawData->region[iRegion].dimAxis[1]; iY++)
        {
            /* Calculate the line position on detector and the start pixel index
             * in bad pixel map */
            line  = (rawData->region[iRegion].corner[1] - 1) + iY;
            
            /* Skip the first amdlibSTART_MASK_COL pixels of each line. These
             * pixels are too noisy, and are not significant when computing
             * bias.  We start by finding the median on 9 values for each
             * detector scanline (row) to avoid spurious spikes, then flag
             * all pixels of the line whose value is five times the
             * median, then average the good pixels left.
             */
            nbGoodPixelsInRow[line] = 0;
            /* skip START_MASK_COL pixels */
            for (iX = amdlibSTART_MASK_COL; 
                 iX < rawData->region[iRegion].dimAxis[0]; 
                 iX++)
            {
                if ((badPixels[iY][iX] == amdlibGOOD_PIXEL_FLAG))
                {
                    sortArray[nbGoodPixelsInRow[line]++] = 
                        rawPtr[iY][iX];
                }
            }
            if (nbGoodPixelsInRow[line] > 8)
            {
                median = amdlibMedian9(sortArray);
            }
            else if (nbGoodPixelsInRow[line] > 0)
            {
                sum = 0.0;
                for (iX = 0; iX < nbGoodPixelsInRow[line] ; iX++)
                {
                    sum += sortArray[iX];
                }
                median = sum / (float)nbGoodPixelsInRow[line];
            }
            else
            {
                median = 0.0;
            }
            rowBias[line] = median;
            /* Redo */
            for (iX = amdlibSTART_MASK_COL; 
                 iX < rawData->region[iRegion].dimAxis[0]; 
                 iX++)
            {
                if ((badPixels[iY][iX] == amdlibGOOD_PIXEL_FLAG))
                {
                    *nbGoodPixelsOnDetector += 1;
                    rmsBias += amdlibPow2(rawPtr[iY][iX] - median);
                }
            }
        }
        amdlibFree2DArrayFloat(badPixels);
        amdlibFree2DArrayFloatWrapping(rawPtr);
    }
    if (*nbGoodPixelsOnDetector != 0)
    {
        rmsBias /= (float)*nbGoodPixelsOnDetector;
    }
    else
    {
        rmsBias = 0.0;
    }

    /* Redo whole computation, avoiding pixels largely beyond this rmsBias */
    *nbGoodPixelsOnDetector = 0;
    finalRmsBias = 0.0;
    for (iRow = 0; iRow < rawData->nbRows; iRow++)
    {
        /* Take first region of each row corresponding to the shielded area */
        iRegion = iRow * rawData->nbCols;
        regionSize = rawData->region[iRegion].dimAxis[0] *
            rawData->region[iRegion].dimAxis[1];
        rawPtr = amdlibWrap2DArrayFloat(
                        &rawData->region[iRegion].data[regionSize*iFrame], 
                        rawData->region[iRegion].dimAxis[0],
                        rawData->region[iRegion].dimAxis[1],
                        errMsg);
        if (rawPtr == NULL)
        {
            return amdlibFAILURE;
        }
        badPixels = amdlibGetBadPixelMapRegion(
                        rawData->region[iRegion].corner[0] - 1,
                        rawData->region[iRegion].corner[1] - 1,
                        rawData->region[iRegion].dimAxis[0],
                        rawData->region[iRegion].dimAxis[1],
                        errMsg);
        if (badPixels == NULL)
        {
            amdlibFree2DArrayFloatWrapping(rawPtr);
            return amdlibFAILURE;
        }
        for (iY = 0; iY <  rawData->region[iRegion].dimAxis[1]; iY++)
        {
            /* Calculate the line position on detector and the start pixel index
             * in bad pixel map */
            line  = (rawData->region[iRegion].corner[1] - 1) + iY;
            nbGoodPixelsInRow[line] = 0;
            /* skip START_MASK_COL pixels */
            for (iX = amdlibSTART_MASK_COL; 
                 iX < rawData->region[iRegion].dimAxis[0]; iX++)
            {
                /* consider only pixels of the line whose value is less than
                 * five times the median */
                if ((badPixels[iY][iX] == amdlibGOOD_PIXEL_FLAG)
                    && (amdlibPow2(rawPtr[iY][iX]-rowBias[line]) < 5 * rmsBias))
                {
                    *nbGoodPixelsOnDetector += 1;
                    sortArray[nbGoodPixelsInRow[line]++] = 
                        rawPtr[iY][iX];
                }
            }
            if (nbGoodPixelsInRow[line] != 0)
            {
                /* average the good pixels left */
                rowBias[line] = 0.0;
                for (iX = 0; iX < nbGoodPixelsInRow[line] ; iX++)
                {
                    rowBias[line] += sortArray[iX];
                }
                rowBias[line] /= (float)(nbGoodPixelsInRow[line]);
                for (iX = 0; iX < nbGoodPixelsInRow[line] ; iX++)
                {
                    finalRmsBias += amdlibPow2(sortArray[iX]-rowBias[line]);
                }
            }
        }
        amdlibFree2DArrayFloat(badPixels);
        amdlibFree2DArrayFloatWrapping(rawPtr);
    }

    if (*nbGoodPixelsOnDetector != 0)
    {
        finalRmsBias /= (float)*nbGoodPixelsOnDetector;
        finalRmsBias = gain * sqrt(finalRmsBias);
    }
    rawData->rmsBias[iFrame] = finalRmsBias;

    return amdlibSUCCESS;
}

/**
 * Calibrate image 
 *
 * This function calibrates the specified image. 
 *
 * @param rawData raw data. 
 * @param iFrame index of the frame 
 * @param gain gain to convert photoelectron counts
 * @param rowBias array where computed bias for each line is stored
 * @param nbSaturatedPixels number of found saturated pixels
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibCalibrateImage(amdlibRAW_DATA        *rawData,
                                      int                   iFrame,
                                      float                 gain,
                                      float                 *rowBias,
                                      int                   *nbSaturatedPixels,
                                      amdlibERROR_MSG       errMsg)
{
    int   iRow, iCol;
    int   iRegion;
    int   iX, iY;
    int   startPixelX, startPixelY;
    int   regionWidth, regionHeight;
    int   regionSize;
    float **badPixels;
    float **flatField;
    float **rawPtr;
    float ffmValue;

    amdlibLogTrace("amdlibCalibrateImage()");
    
    *nbSaturatedPixels = 0;
    for (iRow = 0; iRow < rawData->nbRows; iRow++)
    {
        for (iCol = 0; iCol < rawData->nbCols; iCol++)
        {
            iRegion = iRow * rawData->nbCols + iCol;
            startPixelX  = rawData->region[iRegion].corner[0] - 1;
            regionWidth = rawData->region[iRegion].dimAxis[0];
            startPixelY  = rawData->region[iRegion].corner[1] - 1;
            regionHeight = rawData->region[iRegion].dimAxis[1];

            /* Get bad pixels and flat-field for this region */
            badPixels = amdlibGetBadPixelMapRegion(startPixelX, startPixelY,
                                                   regionWidth, regionHeight,
                                                   errMsg);
            if (badPixels == NULL)
            {
                return amdlibFAILURE;
            }
            flatField = amdlibGetFlatFieldMapRegion(startPixelX, startPixelY,
                                                    regionWidth, regionHeight,
                                                    errMsg);
            if (flatField == NULL)
            {
                amdlibFree2DArrayFloat(badPixels);                
                return amdlibFAILURE;            
            }
            regionSize = regionWidth * regionHeight;
            rawPtr = amdlibWrap2DArrayFloat(
                             &rawData->region[iRegion].data[regionSize*iFrame], 
                             rawData->region[iRegion].dimAxis[0],
                             rawData->region[iRegion].dimAxis[1],
                             errMsg);
            if (rawPtr == NULL)
            {
                amdlibFree2DArrayFloat(badPixels);
                amdlibFree2DArrayFloat(flatField);
                return amdlibFAILURE;
            }
            for (iY = 0; iY < regionHeight; iY++)
            {
                for (iX = 0; iX < regionWidth; iX++)
                {
                    /* Flat is not corrected in shielded area */
                    if (iCol != 0)
                    {
                        ffmValue = flatField[iY][iX];
                    }
                    else
                    {
                        ffmValue = 1.0;
                    }

                    if (badPixels[iY][iX] == amdlibGOOD_PIXEL_FLAG)
                    {
                        /* test wether 'saturated' flag should be set */
                        if (rawPtr[iY][iX] > amdlibDETECTOR_SATURATION) 
                        {
                            *nbSaturatedPixels += 1;
                        }
                        rawPtr[iY][iX] = 
                            gain * (rawPtr[iY][iX] - 
                                    rowBias[startPixelY + iY]) /ffmValue;
                    }
                    else
                    {
                       rawPtr[iY][iX] = amdlibBAD_PIXEL_VALUE;
                    }
                }
            }

            amdlibFree2DArrayFloat(badPixels);
            amdlibFree2DArrayFloatWrapping(rawPtr);
            amdlibFree2DArrayFloat(flatField);
        }
    } 
    return amdlibSUCCESS;
}

/**
 * Subtract remanence
 *
 * This function substracts remanence on all frame. The remanence is computed
 * using amdlibREMANENCE value.
 *
 * @param rawData raw data. 
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibSubtractRemanence(amdlibRAW_DATA  *rawData,
                                         amdlibERROR_MSG errMsg)
{
    int   iRegion;
    int   iFrame;
    int   iX, iY;
    float **memoryOfPreviousImage;
    float ***rawPtr;
    float cal;
    
    amdlibLogTrace("amdlibSubtractRemanence()");
    
    for (iRegion = 0; iRegion < rawData->nbRegions; iRegion++)
    {
        /* Allocate Previous Image area */
        memoryOfPreviousImage = amdlibAlloc2DArrayFloat(
                                        rawData->region[iRegion].dimAxis[0],
                                        rawData->region[iRegion].dimAxis[1],
                                        errMsg);
        if (memoryOfPreviousImage == NULL)
        {
            return amdlibFAILURE;
        }
        rawPtr = amdlibWrap3DArrayFloat(rawData->region[iRegion].data, 
                                        rawData->region[iRegion].dimAxis[0],
                                        rawData->region[iRegion].dimAxis[1],
                                        rawData->region[iRegion].dimAxis[2],
                                        errMsg);
        if (rawPtr == NULL)
        {
            amdlibFree2DArrayFloat(memoryOfPreviousImage);
            return amdlibFAILURE;
        }
        for (iFrame = 0; iFrame < rawData->region[iRegion].dimAxis[2]; iFrame++)
        {
            for (iY = 0; iY < rawData->region[iRegion].dimAxis[1]; iY++)
            {
                for (iX = 0; iX < rawData->region[iRegion].dimAxis[0]; iX++)
                {
                    cal = rawPtr[iFrame][iY][iX] - amdlibREMANENCE * 
                                                memoryOfPreviousImage[iY][iX];
                    memoryOfPreviousImage[iY][iX] = rawPtr[iFrame][iY][iX];
                    rawPtr[iFrame][iY][iX] = cal;
                }
            }
        }
        /*Free Previous Image area */
        amdlibFree2DArrayFloat(memoryOfPreviousImage);
        amdlibFree3DArrayFloatWrapping(rawPtr);
    }
    return amdlibSUCCESS;
}
/*___oOo___*/
